import sys
import argparse
import pysam

basecolor = {'T':'255,0,0', 'A':'0,255,0', 'C':'0,0,255', 'G':'255,165,0'}

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Print a bed file of positions of mononucleotide repeats within an assembly"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 1.0.0"
    )
    parser.add_argument('-r', '--region', required=False, help='genomic region to write bed file for (e.g., chr3:25708893-42710023)')
    parser.add_argument('-f', '--fasta', type=str, required=True, help='fasta file for assembly to proecess')
    parser.add_argument('-m', '--minsize', type=int, required=False, default=10, help='minimum size of mononucleotide repeat to include')
    parser.add_argument('-p', '--prefix', type=str, required=False, default="MNR", help='prefix to use in the names of runs included in the bed file')
    return parser

def find_mononuc_runs(args)->int:
    fastafile = args.fasta
    minsize = args.minsize
    prefix = args.prefix
    refobj = pysam.FastaFile(fastafile)

    refseqs = refobj.references
    repeatnum = 0
    for entry in refseqs:
        fullseq = refobj.fetch(entry).upper()
        curindex = 0
        curstart = 0
        curbase = ""
        for base in fullseq:
            curindex = curindex + 1
            if base != curbase:
                if curindex - curstart - 1 >= minsize and curbase != "N":
                    rgbcol = basecolor[curbase]
                    repeatnum = repeatnum + 1
                    print(entry + "\t" + str(curstart) + "\t" + str(curindex - 1) + "\t" + prefix + "_" + str(repeatnum) + "_" + str(curbase) + "\t1000\t+\t" + str(curstart) + "\t" + str(curindex - 1) + "\t" + rgbcol)
                curbase = base
                curstart = curindex - 1
        if curindex - curstart >= minsize and curbase != "N":
            rgbcol = basecolor[curbase]
            repeatnum = repeatnum + 1
            print(entry + "\t" + str(curstart) + "\t" + str(curindex) + "\t" + prefix + "_" + str(repeatnum) + "_" + str(curbase) + "\t1000\t+\t" + str(curstart) + "\t" + str(curindex - 1) + "\t" + rgbcol)

    return 0

def main() -> None:
    parser = init_argparse()
    args = parser.parse_args()

    find_mononuc_runs(args)

if __name__ == "__main__":
    main()



import shutil
from pathlib import Path
import logging

logger = logging.getLogger(__name__)

def create_output_directory(directory)->None:
    #directory = args.prefix
    path = Path(directory)
    logger.info("Creating directory " + directory + " for output")
    path.mkdir(exist_ok=True)

    return path.as_posix()

def name_output_files(args, outputdir:str)->dict:
    files = {}
    files["nonincludedbed"] = outputdir + "/nonincludedregions." + args.benchmark + ".bed"
    files["allexcludedbed"] = outputdir + "/excludedregions." + args.benchmark + ".bed"
    files["includednonexcludedbed"] = outputdir + "/includednonexcludedregions." + args.benchmark + ".bed"
    files["alignplotdir"] = outputdir + "/alignmentplots"
    files["alignplotprefix"] = outputdir + "/alignmentplots/" + args.assembly + ".clustered_aligns"
    files["testgenomebed"] = outputdir + "/genome." + args.assembly + ".bed"
    if not args.n_bedfile:
        files["testnbed"] = outputdir + "/nlocs." + args.assembly + ".bed"
    else:
        files["testnbed"] = None
    files["testnonnbed"] = outputdir + "/atgcseq." + args.assembly + ".bed"
    files["benchnbed"] = outputdir + "/nlocs." + args.benchmark + ".bed"
    files["benchnonnbed"] = outputdir + "/atgcseq." + args.benchmark + ".bed"
    files["benchgenomebed"] = outputdir + "/genome." + args.benchmark + ".bed"
    files["phasemarkerbed"] = outputdir + "/" + args.assembly + "." + args.benchmark + ".hapmers.bed"
    files["mergedphasemarkerbed"] = outputdir + "/" + args.assembly + "." + args.benchmark + ".hapmers.merged.bed"
    files["mergedmarkerbedwithscaffnames"] = outputdir + "/" + args.assembly + "." + args.benchmark + ".scaffnames.hapmers.merged.bed"
    files["phaseblockbed"] = outputdir + "/" + args.assembly + "." + args.benchmark + ".phasedscaffolds.bed"
    files["hmmphaseblockbed"] = outputdir + "/" + args.assembly + "." + args.benchmark + ".hmmphasedscaffolds.bed"
    files["mergedhmmphaseblockbed"] = outputdir + "/" + args.assembly + "." + args.benchmark + ".merged.hmmphasedscaffolds.bed"
    files["trimmedphasedalignprefix"] = outputdir + "/" + args.assembly + "_vs_" + args.benchmark + ".trimmedphased"
    files["aligntomatbenchprefix"] = args.assembly + "_vs_" + args.benchmark + ".mat"
    files["aligntopatbenchprefix"] = args.assembly + "_vs_" + args.benchmark + ".pat"
    files["matalignedregions"] = outputdir + "/" + args.benchmark + ".mat.covered." + args.assembly + ".bed"
    files["patalignedregions"] = outputdir + "/" + args.benchmark + ".pat.covered." + args.assembly + ".bed"
    files["truthcovered"] = outputdir + "/" + args.assembly + ".benchcovered." + args.benchmark + ".bed"
    files["includedtruthcovered"] = outputdir + "/" + args.assembly + ".includedbenchcovered." + args.benchmark + ".bed"
    files["testmatcovered"] = outputdir + "/testmatcovered." + args.assembly + ".bed"
    files["testpatcovered"] = outputdir + "/testpatcovered." + args.assembly + ".bed"
    files["generalstatsfile"] = outputdir + "/" + args.assembly + ".generalstats.txt"
    files["contiglengths"] = outputdir + "/" + args.assembly + ".contiglengths.txt"
    files["scaffoldlengths"] = outputdir + "/" + args.assembly + ".scaffoldlengths.txt"
    files["mononucstatsfile"] = outputdir + "/" + args.assembly + ".mononucstats.txt"
    files["structdetailsfile"] = outputdir + "/" + args.assembly + ".structurestats.txt"
    files["structvariantbed"] = outputdir + "/" + args.assembly + ".svs.bed"
    files["structvariantsvcf"] = outputdir + "/" + args.assembly + ".svs.vcf"
    files["clusterlengths"] = outputdir + "/" + args.assembly + ".alignclusterlengths.txt"
    files["coveredmononucsfile"] = outputdir + "/" + args.assembly + ".coveredmononucs." + args.benchmark + ".bed"
    files["coveredincludedmononucsfile"] = outputdir + "/" + args.assembly + ".coveredincludedmononucs." + args.benchmark + ".bed"
    files["mononucswithvariantsfile"] = outputdir + "/" + args.assembly + ".mononucswithvariants." + args.benchmark + ".bed"
    files["bencherrortypebed"] = outputdir + "/" + args.assembly + ".errortype." + args.benchmark + ".bed"
    files["benchexcludederrortypebed"] = outputdir + "/" + args.assembly + ".excludederrors." + args.benchmark + ".bed"
    files["testerrortypebed"] = outputdir + "/errortype." + args.assembly + ".bed"
    files["coveredhetsitealleles"] = outputdir + "/" + args.benchmark + ".coveredhetalleles." + args.assembly + ".bed"
    files["snvstatsfile"] = outputdir + "/" + args.assembly + ".singlenucerrorstats.txt"
    files["indelstatsfile"] = outputdir + "/" + args.assembly + ".indelerrorstats.txt"
    files["qvstatsfile"] = outputdir + "/" + args.assembly + ".qvstats.txt"

    return files

def name_read_stats_files(args, outputdir:str)->dict:
    files = {}
    files["includedbedfile"] = outputdir + "/" + args.readsetname + ".includedregions." + args.benchmark + ".bed"
    files["mononucstatsfile"] = outputdir + "/" + args.readsetname + ".mononucstats.txt"
    files["mononuchistfile"] = outputdir + "/" + args.readsetname + ".mononuchist.txt"
    files["mononuccompositionfile"] = outputdir + "/" + args.readsetname + ".mononuccomposition.txt"
    files["includedmononucfile"] = outputdir + "/" + args.readsetname + ".includedmononucs.txt"
    files["mononucoverviewfile"] = outputdir + "/" + args.readsetname + ".mononucgeneralstats.txt"
    files["includedstrprefix"] = outputdir + "/" + args.readsetname + ".includedstrs"
    files["strstatsprefix"] = outputdir + "/" + args.readsetname + ".strstats"
    files["strlengthdiffaccuracyprefix"] = outputdir + "/" + args.readsetname + ".strlengthdiffaccuracy"
    files["strlengthaccuracyprefix"] = outputdir + "/" + args.readsetname + ".strlengthaccuracy"
    files["readerrorfile"] = outputdir + "/" + args.readsetname + ".readerrors.txt"
    files["errorstatsfile"] = outputdir + "/" + args.readsetname + ".generalstats.txt"
    files["snvstatsfile"] = outputdir + "/" + args.readsetname + ".singlenucerrorstats.txt"
    files["indelstatsfile"] = outputdir + "/" + args.readsetname + ".indelerrorstats.txt"
    files["qvstatsfile"] = outputdir + "/" + args.readsetname + ".errorqvstats.txt"
    files["coveragebedfile"] = outputdir + "/" + args.readsetname + ".binnedcoverage.bed"
    files["arrivalratebedfile"] = outputdir + "/" + args.readsetname + ".arrivalrates.bed"
    files["includedcoveragebedfile"] = outputdir + "/" + args.readsetname + ".binnedcoverage.included.bed"
    files["extremekmersbedfile"] = outputdir + "/" + args.readsetname + ".extremekmercounts.bed"
    files["benchmarkkmercountfile"] = outputdir + "/" + args.readsetname + ".benchmarkkmercounts.txt"
    files["readalignedkmercountfile"] = outputdir + "/" + args.readsetname + ".readalignedkmercounts.txt"
    files["strandedalignedkmercountfile"] = outputdir + "/" + args.readsetname + ".strandedalignedkmercounts.txt"

    return files

def name_constructgenome_output_files(args, outputdir:str)->dict:
    files = {}
    prefix = args.prefix
    files["hap1vcf"] = outputdir + "/" + prefix + ".hap1.vcf.gz"
    files["hap2vcf"] = outputdir + "/" + prefix + ".hap2.vcf.gz"
    files["hap1chain"] = outputdir + "/" + prefix + ".hap1.chain"
    files["hap2chain"] = outputdir + "/" + prefix + ".hap2.chain"
    files["hap1mutfasta"] = outputdir + "/" + prefix + ".hap1.mut.fasta"
    files["hap2mutfasta"] = outputdir + "/" + prefix + ".hap2.mut.fasta"
    files["hap1hcbed"] = outputdir + "/" + prefix + ".hap1.highconf.bed"
    files["hap2hcbed"] = outputdir + "/" + prefix + ".hap2.highconf.bed"
    files["hap1unmapped"] = outputdir + "/" + prefix + ".hap1.unmapped"
    files["hap2unmapped"] = outputdir + "/" + prefix + ".hap2.unmapped"
    files["hap1lowqualbed"] = outputdir + "/" + prefix + ".hap1.lowconf.bed"
    files["hap2lowqualbed"] = outputdir + "/" + prefix + ".hap2.lowconf.bed"
    files["hap1maskedfasta"] = outputdir + "/" + prefix + ".hap1.masked.fasta"
    files["hap2maskedfasta"] = outputdir + "/" + prefix + ".hap2.masked.fasta"

    return files

def name_compare_output_files(args, outputdir:str)->dict:
    files = {}
    #files[""] = outputdir + "/" + args.A + "_vs_" + args.B + ".bed"

    return files




import sys
import os
import re
import shutil
import pysam
import argparse
import logging
from pybedtools import BedTool
import importlib.resources
from pathlib import Path
from collections import namedtuple
from GQC import bedtoolslib
from GQC import errors
from GQC import output
from GQC import seqparse
from GQC import align
from GQC import alignparse
from GQC import structvar
from GQC import phasing
from GQC import stats
from GQC import mummermethods
from GQC import plots

# create namedtuple for bed intervals:
varianttuple = namedtuple('varianttuple', ['chrom', 'start', 'end', 'name', 'vartype', 'excluded', 'qvscore']) 

logger = logging.getLogger(__name__)

def check_for_bedtools():
    if shutil.which("bedtools") is None:
        print("You don\'t seem to have bedtools in your path. Please install bedtools")
        logger.critical("You don\'t seem to have bedtools in your path. Please install bedtools")
        exit(1)
    return 0

def check_for_aligner(aligner:str):
    if aligner == "winnowmap2":
        aligner = "winnowmap"
    if shutil.which(aligner) is None:
        print("You don\'t seem to have the required aligner " + aligner + " installed in your path. Please install it")
        logger.critical("You don\'t seem to have the required aligner " + aligner + " installed in your path. Please install it")
        exit(1)
    return 0

def check_for_R():
    if shutil.which("Rscript") is None:
        print("You don\'t seem to have Rscript in your path. Plots will not be generated")
        logger.warning("You don\'t seem to have Rscript in your path. Plots will not be generated")
        return 1
    return 0

def check_for_fastk():
    if shutil.which("FastK") is None or shutil.which("KmerMap") is None:
        print("You don\'t seem to have FastK with the KmerMap command in your path. KmerMap is included in the v1.2 release of FASTK. These are necessary so scaffold phase blocks can be assessed")
        logger.critical("You don\'t seem to have FastK with the KmerMap command in your path. KmerMap is included in the v1.2 release of FASTK. These are necessary so scaffold phase blocks can be assessed")
        exit(1)
    return 0

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Print assembly statistics from bam files of the assembly aligned to a benchmark assembly"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('-b', '--bam', required=False, default=None, help='bam file of alignments of the test (haploid) assembly to the diploid benchmark')
    parser.add_argument('--paf', required=False, default=None, help='paf-formatted file of alignments of the test (haploid) assembly to the diploid benchmark')
    parser.add_argument('-r', '--reffasta', type=str, required=True, help='(indexed) fasta file for benchmark reference')
    parser.add_argument('-q', '--queryfasta', type=str, required=True, help='(indexed) fasta file for haploid or diploid test assembly')
    parser.add_argument('-p', '--prefix', type=str, required=True, help='prefix for output directory name, filenames have assembly name in prefix (see -A)')
    parser.add_argument('-t', type=int, required=False, default=2, help='number of processors to use')
    parser.add_argument('-a', '--aligner', type=str, required=False, default='minimap2', help='aligner to use when comparing assembly to benchmark, can be minimap2 or winnowmap2 (default winnowmap2)')
    parser.add_argument('-m', '--minalignlength', type=int, required=False, default=500, help='minimum length of alignment required to be included in alignment statistics and error counts')
    parser.add_argument('--mincontiglength', type=int, required=False, default=500, help='minimum length for contig to be included in contig statistics')
    parser.add_argument('--minns', type=int, required=False, default=10, help='minimum number of consecutive Ns required to break scaffolds into contigs')
    parser.add_argument('--maxclusterdistance', type=int, required=False, default=10000, help='maximum distance within a cluster of alignments')
    parser.add_argument('--merquryblocks', action='store_true', required=False, help='flag option for calculating phase blocks of the test assembly scaffolds using the algorithm used in Merqury rather than an HMM')
    parser.add_argument('--shortnum', type=int, required=False, default=100, help='maximum number of short-range phase switches to allow in phase blocks (if --merquryblocks option is specified)')
    parser.add_argument('--shortlimit', type=int, required=False, default=20000, help='maximum lengths of short-range switches allowable in phase blocks (if --merquryblocks option is specified)')
    parser.add_argument('--includefile', type=str, required=False, default=None, help='bed file of benchmark locations to include in the evaluation (stretches of 10 or more Ns and regions excluded in the exclude file will still not be considered)')
    parser.add_argument('--excludefile', type=str, required=False, default=None, help='bed file of benchmark locations to exclude from consideration (in addition to stretches of 10 or more Ns and regions in the exclude file specified in the config file)')
    parser.add_argument('--vcf', action='store_true', required=False, default=False, help='write differences from benchmark in VCF format')
    parser.add_argument('-n', '--n_bedfile', type=str, required=False, default=None, help='pre-existing bedfile of locations of N-stretches splitting scaffolds into contigs')
    parser.add_argument('--variantfile', type=str, required=False, default=None, help='pre-existing file of variant locations in assembly compared to benchmark')
    parser.add_argument('--structureonly', action='store_true', required=False, help='analyse only the long-range structure of the assembly')
    parser.add_argument('-A', '--assembly', type=str, required=False, default="test", help='name of the assembly being tested--should correspond to query sequence in bam file and will be used in output file names')
    parser.add_argument('-B', '--benchmark', type=str, required=False, default="truth", help='name of the assembly being used as a benchmark--should be the reference sequence in the bam file')
    parser.add_argument('-c', '--config', type=str, required=False, default="benchconfig.txt", help='path to a config file specifying locations of benchmark data files')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging purposes')
    parser.add_argument('--nosplit', action='store_true', required=False, help='use haplotype alignments without splitting on large indels. Default is to split alignments at locations with indels of at least --splitdistance')
    parser.add_argument('--splitdistance', type=int, required=False, default=10000, help='By default, split alignments when they contain indels of this size or greater')
    parser.add_argument('--alpha', type=float, required=False, default=0.05, help='emission probability for displaying opposite haplotype markers in HMM phase block algorithm')
    parser.add_argument('--beta', type=float, required=False, default=0.01, help='transition probability for changing haplotype state between adjacent markers (regardless of distance between them, unless --distancemultiplier value is set) in HMM phase block algorithm')
    #parser.add_argument('--distancemultiplier', type=int, required=False, default=0, help='set the transition probability between two different states to the value of beta times the number of bases between the two markers divided by this value')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def read_config_data(args)->dict:
    configfile = args.config
    configpath = Path(configfile)

    configvals = {}
    if not configpath.exists():
        logger.critical("A config file must exist in the default location benchconfig.txt or be specified with the --config option.")
        print("A config file must exist in the default location benchconfig.txt or be specified with the --config option.")
        exit(1)
    else:
        logger.info("Using resource locations from " + configfile)
        with open(configfile, "r") as cr:
            configline = cr.readline()
            while configline:
                p = re.compile(r'^([^#\s]+):+\s+(\S+)\s*$')
                match = p.match(configline)
                if match:
                    key = match.group(1)
                    value = match.group(2)
                    configvals[key] = value
                configline = cr.readline()

        # add the resource directory location for all non-absolute paths:
        if 'resourcedir' in configvals.keys():
            resourcedir = configvals["resourcedir"]
            if (os.path.exists(resourcedir)):
                for configkey in configvals.keys():
                    if configvals[configkey][0] != "/" and configkey != 'matpattern' and configkey != 'patpattern':
                        configvals[configkey] = resourcedir + "/" + configvals[configkey]
            else:
                logger.critical("The resource directory specified in the config file as \"resourcedir\" does not exist. Please change it to the location of files from the resource tarball")
                print("The resource directory specified in the config file as \"resourcedir\" does not exist. Please change it to the location of files from the resource tarball")
                exit(1)
   
    if 'matpattern' not in configvals.keys():
        configvals['matpattern'] = 'MAT'
    if 'patpattern' not in configvals.keys():
        configvals['patpattern'] = 'PAT'

    return configvals

def main() -> None:

    args = parse_arguments(sys.argv[1:])

    logfile = args.prefix + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output to ' + logfile + ' for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    # log the command line:
    call_command = " ".join(sys.argv)
    logger.info(call_command)

    # check for necessary installed programs and write an output directory:
    check_for_bedtools()
    check_for_aligner(args.aligner)
    no_rscript = check_for_R()

    # dictionary of parameters from the benchmark configuration file:
    benchparams = read_config_data(args)

    # pysam objects for the benchmark and test assembly fasta files:
    ref = Path(args.reffasta)
    query = Path(args.queryfasta)
    if not ref.is_file() or not query.is_file():
        logger.critical("Ref fasta file " + args.reffasta + " and query fasta file " + args.queryfasta + " must exist and be readable")
        exit(1)
    refobj = pysam.FastaFile(args.reffasta)
    queryobj = pysam.FastaFile(args.queryfasta)

    outputdir = output.create_output_directory(args.prefix)

    # dictionary of this run's output file names:
    outputfiles = output.name_output_files(args, outputdir)

    logger.info("Step 1 (of 12): Writing bed files for excluded regions, test assembly and benchmark scaffold spans, lengths, N stretches, and contigs (ignoring stretches of less than " + str(args.minns) + " Ns)")
    bedregiondict = {}
    # merged excluded regions are saved as BedTool object "allexcludedregions" in bedregiondict here
    seqparse.write_genome_bedfiles(queryobj, refobj, args, benchparams, outputfiles, bedregiondict)
   

    # find general stats about contig/scaffold lengths, N/L50's, etc.:
    logger.info("Step 2 (of 12): Writing general statistics about " + args.assembly + " assembly")
    benchmark_stats = stats.write_general_assembly_stats(refobj, queryobj, bedregiondict["testnonnregions"], bedregiondict["testnregions"], outputfiles, benchparams, args)

    logger.info("Step 3 (of 12): Phasing assembly regions using benchmark's haplotype-specific kmers")
    if (not os.path.exists(outputfiles["phaseblockbed"])) or (os.path.getsize(outputfiles["phaseblockbed"]) == 0):
        if args.merquryblocks:
            markerbed = outputfiles["phasemarkerbed"]
        else:
            markerbed = outputfiles["mergedphasemarkerbed"]
        if not os.path.exists(markerbed):
            print("File " + markerbed + " does not exist--calling FASTK to generate one")
            logger.info("File " + markerbed + " does not exist--calling FASTK to generate one")
            check_for_fastk()
            phasing.map_benchmark_hapmers_onto_assembly(args.queryfasta, benchparams["matmarkerdb"], benchparams["patmarkerdb"], outputdir, outputfiles)
    
        if args.merquryblocks: # use Merqury phase block algorithm
            logger.info("Using Merqury algorithm with shortnum " + str(args.shortnum) + " and shortlimit " + str(args.shortlimit) + " to find phase blocks")
            shortnum = args.shortnum
            shortlimit = args.shortlimit
            matpattern = benchparams["matpattern"]
            phaseblockints = phasing.find_phase_blocks_from_marker_bed(markerbed, queryobj.references, shortnum, shortlimit, matpattern)
        else: # use HMM algorithm to find phase blocks
            logger.info("Using HMM with emission probability " + str(args.alpha) + " and transition probability " + str(args.beta) + " to find phase blocks")
            alpha = args.alpha
            transitionprob = args.beta
            phaseblockints = phasing.find_hapmer_phase_blocks_with_hmm(markerbed, outputfiles["hmmphaseblockbed"], queryobj, alpha, transitionprob, 0)
        phaseblockints.saveas(outputfiles["phaseblockbed"])
        matphaseblockints = phaseblockints.filter(lambda x: x.name=="mat")
        patphaseblockints = phaseblockints.filter(lambda x: x.name=="pat")
    else:
        logger.info("Skipping assembly phasing as file " + outputfiles["phaseblockbed"] + " already exists and is not empty")
        phaseblockints = BedTool(outputfiles["phaseblockbed"])
        matphaseblockints = phaseblockints.filter(lambda x: x.name=="mat")
        patphaseblockints = phaseblockints.filter(lambda x: x.name=="pat")

    #stats.write_phase_block_stats(phaseblockints, outputfiles, benchmark_stats, args)

    # align test assembly separately to maternal and paternal haplotypes:
    logger.info("Step 4 (of 12): Aligning assembly separately to maternal and paternal haplotypes of the benchmark and gathering trimmed alignments of phased assembly regions to their corresponding haplotypes")

    trimmedphasedbam = outputfiles["trimmedphasedalignprefix"] + ".merge.sort.bam"
    if not os.path.exists(trimmedphasedbam):
        logger.debug("Creating trimmed sorted bam " + trimmedphasedbam)
        [matbenchbamfile, patbenchbamfile] = align.align_assembly_to_benchmark_haplotypes(args.queryfasta, outputfiles, benchparams, args)
    
        # read in alignments from BAM format, filtering out secondaries and finding the optimal alignment on the correct haplotype for each phase block in the assembly

        benchdiploidheaderstring = "@HD\tVN:1.6\tSO:coordinate\n"
        for refentry in refobj.references:
            entrylength = refobj.get_reference_length(refentry)
            benchdiploidheaderstring += "@SQ\tSN:" + refentry + "\tLN:"  + str(entrylength) + "\n"
    
        [mataligns, matalignedintervals] = alignparse.index_aligns_by_boundaries(matbenchbamfile, args)
        [pataligns, patalignedintervals] = alignparse.index_aligns_by_boundaries(patbenchbamfile, args)
    
        matalignedintervals.saveas(outputfiles["matalignedregions"])
        patalignedintervals.saveas(outputfiles["patalignedregions"])
    
        print("Finding subaligns in maternal alignments for maternal phase blocked regions of the assembly")
        matblocksubaligns = alignparse.find_phaseblock_subaligns(matphaseblockints, matalignedintervals, mataligns)
        mattrimmedbamfile = outputfiles["trimmedphasedalignprefix"] + ".mat.bam"
        alignparse.write_aligns_to_bamfile(mattrimmedbamfile, matblocksubaligns, headerbam=matbenchbamfile, sort=False)
        print("Finding subaligns in paternal alignments for paternal phase blocked regions of the assembly")
        patblocksubaligns = alignparse.find_phaseblock_subaligns(patphaseblockints, patalignedintervals, pataligns)
        pattrimmedbamfile = outputfiles["trimmedphasedalignprefix"] + ".pat.bam"
        alignparse.write_aligns_to_bamfile(pattrimmedbamfile, patblocksubaligns, headerbam=patbenchbamfile, sort=False)

        # now merge the maternal and paternal trimmed files to a single file with a diploid header, sort, and index:
        alignparse.merge_trimmed_bamfiles(mattrimmedbamfile, pattrimmedbamfile, benchdiploidheaderstring, outputfiles)
    else: 
        logger.info("Skipping step 4 (of 12): Trimmed phased alignments already exist in " + trimmedphasedbam)

    logger.info("Step 5 (of 12): Filtering alignment to include primary best increasing subset")
    if not args.nosplit:
        logger.info("Splitting alignments at indels greater than or equal to " + str(args.splitdistance) + " bases")
        splitbam_name = trimmedphasedbam.replace(".bam", ".split.bam")
        splitsortbam_name = splitbam_name.replace(".bam", ".sort.bam")
        if not os.path.exists(splitsortbam_name):
            alignobj = pysam.AlignmentFile(trimmedphasedbam, "rb")
            alignparse.split_aligns_and_sort(splitbam_name, alignobj, minindelsize=args.splitdistance)
            pysam.sort("-o", splitsortbam_name, splitbam_name)
            pysam.index(splitsortbam_name)
            os.remove(splitbam_name)
        alignobj = pysam.AlignmentFile(splitsortbam_name, "rb")
        aligndata = alignparse.read_bam_aligns(alignobj, args.minalignlength)
        rlis_aligndata = mummermethods.filter_aligns(aligndata, "target")
    else:
        logger.info("Not splitting alignments--using alignments at least " + str(args.minalignlength) + " bases from " + trimmedphasedbam)
        alignobj = pysam.AlignmentFile(trimmedphasedbam, "rb")
        aligndata = alignparse.read_bam_aligns(alignobj, args.minalignlength)
        rlis_aligndata = mummermethods.filter_aligns(aligndata, "target")

    ## find clusters of consistent, covering alignments and calculate continuity statistics:
    logger.info("Step 6 (of 12): Assessing overall structural alignment of assembly")
    # (by default, rlis_aligndata are split alignments filtered for RLIS)
    if not os.path.exists(outputfiles["clusterlengths"]) or not os.path.exists(outputfiles["alignplotdir"]):
        alignparse.assess_overall_structure(rlis_aligndata, refobj, queryobj, outputfiles, bedregiondict, benchmark_stats, args)
        structvar.write_structural_errors(rlis_aligndata, refobj, queryobj, outputfiles, benchmark_stats, args)
        stats.write_aligned_cluster_stats(outputfiles, benchmark_stats, args)
    else:
        logger.info("Skipping overall structural assessment because " + outputfiles["clusterlengths"] + " and directory " + outputfiles["alignplotdir"] + " already exist")

    if not args.structureonly:
       logger.info("Step 7 (of 12): Writing bed files of regions covered by alignments of " + args.assembly + " to " + args.benchmark + " (includes phasing of assembly at benchmark heterozygous sites)")
       ## read in locations of het variants in the benchmark store in dictionary by chrom of position-sorted arrays:
       hetsites = phasing.read_hetsites(benchparams["hetsitevariants"])
       hetarrays = phasing.sort_chrom_hetsite_arrays(hetsites)

       pafaligns = None
       [refcoveredbed, querycoveredbed, variants, hetsitealleles, alignedscorecounts, snverrorscorecounts, indelerrorscorecounts] = alignparse.write_bedfiles(alignobj, pafaligns, refobj, queryobj, hetarrays, outputfiles["testmatcovered"], outputfiles["testpatcovered"], outputfiles["truthcovered"], outputfiles["coveredhetsitealleles"], bedregiondict["allexcludedregions"], benchparams, args)

       ## create non-excluded covered file:
       bedtoolslib.intersectbed(outputfiles["truthcovered"], outputfiles["includednonexcludedbed"], outputfile=outputfiles["includedtruthcovered"])

       ## create merged unique outputfiles:
       [mergedtruthcoveredbed, outputfiles["mergedtruthcovered"]] = bedtoolslib.mergebed(outputfiles["truthcovered"])
       [mergedincludedtruthcoveredbed, outputfiles["mergedincludedtruthcovered"]] = bedtoolslib.mergebed(outputfiles["includedtruthcovered"])
       [mergedtestmatcoveredbed, outputfiles["mergedtestmatcovered"]] = bedtoolslib.mergebed(outputfiles["testmatcovered"])
       [mergedtestpatcoveredbed, outputfiles["mergedtestpatcovered"]] = bedtoolslib.mergebed(outputfiles["testpatcovered"])

       logger.info("Step 8 (of 12): Writing primary alignment statistics about " + args.assembly + " assembly")
       stats.write_merged_aligned_stats(refobj, queryobj, mergedtruthcoveredbed, mergedincludedtruthcoveredbed, mergedtestmatcoveredbed, mergedtestpatcoveredbed, outputfiles, benchmark_stats, benchparams, args)

       if alignobj is not None:
           ## classify variant errors as phasing or novel errors:
           logger.info("Step 9 (of 12): Writing phase switch statistics to generalstats.txt file")
           stats.write_het_stats(outputfiles, benchmark_stats, benchparams, args)
           stats.write_hmm_phaseswitch_stats(outputfiles, benchmark_stats, args)
           logger.info("Step 10 (of 12): Determining whether errors are switched haplotype or novel")
           errors.classify_errors(refobj, queryobj, variants, hetsites, outputfiles, benchparams, benchmark_stats, args)
           stats.write_qv_stats(benchmark_stats, alignedscorecounts, snverrorscorecounts, indelerrorscorecounts, outputfiles, args)
#
           ## evaluate mononucleotide runs:
           logger.info("Step 11 (of 12): Assessing accuracy of mononucleotide runs")
           bedtoolslib.intersectbed(benchparams["mononucruns"], outputfiles["mergedtruthcovered"], outputfile=outputfiles["coveredmononucsfile"], writefirst=True)
           bedtoolslib.intersectbed(outputfiles["coveredmononucsfile"], outputfiles["allexcludedbed"], outputfile=outputfiles["coveredincludedmononucsfile"], writefirst=True, v=True)
           [mononucswithvariantsbed, mononucswithvariantsbedfile] = bedtoolslib.intersectbed(outputfiles["coveredincludedmononucsfile"], outputfiles["bencherrortypebed"], outputfiles["mononucswithvariantsfile"], outerjoin=True, writeboth=True)
           mononucstats = errors.gather_mononuc_stats(outputfiles["mononucswithvariantsfile"], outputfiles["mononucstatsfile"])
           stats.write_mononuc_stats(mononucstats, outputfiles, benchmark_stats, args)
    else:
        logger.info("Skipping steps 7 through 11 (of 12) since --structureonly option was used")

    # plot alignment coverage across assembly and genome:
    if not no_rscript:
        logger.info("Step 12 (of 12): Creating plots")
        if not args.structureonly:
            plots.plot_benchmark_align_coverage(args.assembly, args.benchmark, outputdir, outputfiles["benchgenomebed"], outputfiles["benchnbed"], benchparams["matpattern"])
            plots.plot_testassembly_align_coverage(args.assembly, args.benchmark, outputdir, benchparams["resourcedir"])
            plots.plot_assembly_error_stats(args.assembly, args.benchmark, outputdir)
            if alignobj is not None:
                plots.plot_mononuc_accuracy(args.assembly, args.benchmark, outputdir, benchparams["resourcedir"])
                if len(alignedscorecounts) > 0:
                    plots.plot_qv_score_concordance(args.assembly, args.benchmark, outputdir, benchparams["resourcedir"])
            if os.path.getsize(mononucswithvariantsbedfile) > 0:
                plots.plot_mononuc_accuracy(args.assembly, args.benchmark, outputdir, benchparams["resourcedir"])
            else:
                logger.info("No mononucleotide runs assessed--skipping plot")
            plots.plot_assembly_error_stats(args.assembly, args.benchmark, outputdir)
            plots.plot_assembly_discrepancy_counts(args.assembly, args.benchmark,  outputdir)
            plots.plot_svcluster_align_plots(args.assembly, args.benchmark, outputfiles["alignplotdir"], refobj, mode='bench')
        plots.run_ngax_plot(args.assembly, args.benchmark, outputdir, outputfiles["benchgenomebed"], benchparams["resourcedir"])
        if "assemblyqv" in benchmark_stats.keys():
            assemblyqv = benchmark_stats["assemblyqv"]
        else:
            assemblyqv = "NA"
        plots.plot_assembly_summary_stats(args.assembly, args.benchmark,  outputdir, outputfiles["benchgenomebed"], benchparams["resourcedir"], assemblyqv=assemblyqv)


if __name__ == "__main__":
    main()



import sys
import re
import logging
import igraph as ig

logger = logging.getLogger(__name__)

def read_node_file(filename):
    nodes = {}

    with open(filename) as fh:
        for line in fh.readlines():
            fields = line.rstrip().split()
            nodes[fields[0]] = 1

    return nodes

def read_graph_from_gfa(gfafile, args):

    allnodes = {}
    nodelines = {}
    edgelines = []
    nodelengths = {}
    with open(gfafile, "r") as gfh:
        for line in gfh.readlines():
            fields = line.rstrip().split()
            if fields[0] == 'S':
                nodename = fields[1]
                lengthmatch = re.search('(?<=LN:i:)\d+$', fields[3])
                if lengthmatch:
                    nodelengths[nodename] = int(lengthmatch.group(0))
                nodelines[nodename] = line
            elif fields[0] == 'L':
                edgelines.append(line)
                allnodes[fields[1]] = 1
                allnodes[fields[3]] = 1

    #construct graph:
    n_vertices = len(allnodes.keys())
    node_indices = {}
    nodes = []
    edges = []
    # directions will be ">" for "-" nodes and "<" for "+" nodes
    directionlist = []
    index = 0
    for node in allnodes.keys():
        node_indices[node] = index
        nodes.append(node)

        #print("Node " + node + " has length " + str(parednodelengths[node]))
        index = index + 1
    for line in edgelines:
        fields = line.rstrip().split()
        node1 = fields[1]
        dir1 = fields[2]
        node2 = fields[3]
        dir2 = fields[4]
        edges.append([node_indices[node1], node_indices[node2]])
        directionlist.append(dir1 + dir2)

    assemblygraph = ig.Graph(n=n_vertices, edges=edges, directed=True)
    assemblygraph.vs["name"] = nodes
    assemblygraph.es["direction"] = directionlist

    return assemblygraph

def find_hom_het_nodes(graph, args):

    # look for nodes with two incoming or two outgoing edges:
    nodelist = graph.vs
    nodenamelist = nodelist["name"]
    edgelist = graph.es
    for node in graph.vs:
        attribdict = node.attributes()
        nodename = attribdict["name"]
        #print(nodename)
        nodeedges = node.out_edges()
        numedges = len(nodeedges)
        #print("Node " + nodename + " has " + str(numedges) + " outgoing edges:")
        numbackedges = 0
        numforwardedges = 0
        backedges = []
        forwardedges = []
        for edge in nodeedges:
            node1 = nodenamelist[edge.tuple[0]]
            node2 = nodenamelist[edge.tuple[1]]
            node1dir = edge.attributes()["direction"][0]
            node2dir = edge.attributes()["direction"][1]
            if node1dir == "-":
                numbackedges = numbackedges + 1
                backedges.append(edge)
            elif node1dir == "+":
                numforwardedges = numforwardedges + 1
                forwardedges.append(edge)
            #print(str(edge.tuple) + " " +  node1 + " " + node2 + " " + edge.attributes()["direction"])
        if numbackedges == 2:
            # trace back to a hom node if possible:
            # make a list of nodes extending from the 5' end of the unitig, as well as nodes at ends of edges at ends of single edges from the opposite ends:
            backhetnodes = []
            backhetsecondnodes = []
            for backedge in backedges:
                node2dir = backedge.attributes()["direction"][1]
                backhetnodes.append(nodenamelist[backedge.tuple[1]])
                doublebackhetedges = nodelist[backedge.tuple[1]].out_edges()
                extendingdoublebackhetedges = []
                for hetedge in doublebackhetedges:
                    if hetedge.attributes()["direction"][0] == node2dir:
                        extendingdoublebackhetedges.append(hetedge)
                numdoublebackhetedges = len(extendingdoublebackhetedges)
                if numdoublebackhetedges == 1:
                    backhetsecondnodes.append(nodenamelist[extendingdoublebackhetedges[0].tuple[1]])
            if len(backhetsecondnodes)==2 and backhetsecondnodes[0]==backhetsecondnodes[1]:
                print("Found backward hets from " + node1 + " named " + backhetnodes[0] + " and " + backhetnodes[1] + " leading to " + backhetsecondnodes[0])
        if numforwardedges == 2:
            # trace fwd to a hom node if possible:
            # make a list of nodes extending from the 3' end of the unitig, as well as nodes at ends of edges at ends of single edges from the opposite ends:
            fwdhetnodes = []
            fwdhetsecondnodes = []
            for fwdedge in forwardedges:
                node2dir = fwdedge.attributes()["direction"][1]
                fwdhetnodes.append(nodenamelist[fwdedge.tuple[1]])
                doublefwdhetedges = nodelist[fwdedge.tuple[1]].out_edges()
                extendingdoublefwdhetedges = []
                for hetedge in doublefwdhetedges:
                    if hetedge.attributes()["direction"][0] == node2dir:
                        extendingdoublefwdhetedges.append(hetedge)
                numdoublefwdhetedges = len(extendingdoublefwdhetedges)
                if numdoublefwdhetedges == 1:
                    fwdhetsecondnodes.append(nodenamelist[extendingdoublefwdhetedges[0].tuple[1]])
            if len(fwdhetsecondnodes)==2 and fwdhetsecondnodes[0]==fwdhetsecondnodes[1]:
                print("Found forward hets from " + node1 + " named " + fwdhetnodes[0] + " and " + fwdhetnodes[1] + " leading to " + fwdhetsecondnodes[0])

def compare_compressed_fasta_to_node_seqs(fastafile:str, nodegfa:str, args):
    minalignlength = args.minalignlength
    if args.prefix is not None:
        assemblyprefix = args.prefix
        nodefasta = args.prefix + ".nodes.fasta"
        compressedfasta = args.prefix + ".hpc.fasta"
        chainfile = args.prefix + ".hpc.chain"
    else:
        compressedfasta = re.sub("\.fa.*$", ".hpc.fasta", fastafile)
        compressedfasta = re.sub(".*/", "", compressedfasta)
        assemblyprefix = re.sub("\.fa.*$", "", fastafile)
        assemblyprefix = re.sub(".*/", "", assemblyprefix)
        nodefasta = re.sub("\.gfa.*$", ".nodes.fasta", nodegfa)
        nodefasta = re.sub(".*/", "", nodefasta)
        chainfile = re.sub("\.fa.*$", ".hpc.chain", fastafile)
        chainfile = re.sub(".*/", "", chainfile)

    paffile = re.sub("\.fasta$", "_vs_", nodefasta)
    paffile = paffile + assemblyprefix + ".paf"
    hpcbed = re.sub(".paf$", ".sort.bed", paffile)
    decompbed = re.sub(".bed", ".decompressed.bed", hpcbed)
    decompunmapped = re.sub(".bed", ".decompressed.unmapped", hpcbed)

    if os.path.isfile(nodefasta):
        print("Using pre-existing node fasta file " + nodefasta)
    else:
        print("Writing node fasta file " + nodefasta)
        write_gfa_fasta(nodegfa, nodefasta)

    if os.path.isfile(paffile):
        print("Using pre-existing alignment file " + paffile)
    else:
        print("Running minimap2 to generate " + paffile)
        commandwords = ['minimap2', '-k', '28', '-w', '100', '-x', 'lr:hq', '-t', '2', '-r', '10,10', '-N', '1', nodefasta, compressedfasta, '-o', paffile]
        subprocess.run(commandwords)

    bedlines = {} # dictionary to be sorted before printing
    with open(paffile, "r") as pfh:
        for line in pfh:
            fields = line.split()
            compchrom = fields[0]
            compchromlength = int(fields[1])
            compstart = int(fields[2])
            compend = int(fields[3])
            strand = fields[4]
            node = fields[5]
            if strand == "+":
                node = node + "<"
            else:
                node = node + ">"
            nodelength = int(fields[6])
            nodealigned = int(fields[8]) - int(fields[7])
            fracaligned = int(100*nodealigned/nodelength)
            if compend - compstart < minalignlength:
                continue
            compstartstring = str(compstart).zfill(12)
            compendstring = str(compend).zfill(12)
            bedlines[compchrom + ":" + compstartstring + ":" + compendstring] = compchrom + "\t" + str(compstart) + "\t" + str(compend) + "\t" + node + "\t" + str(fracaligned) + "\n"

    with open(hpcbed, "w") as bfh:
        for posstring in sorted(bedlines.keys()):
            bfh.write(bedlines[posstring])

    print("Running liftover to decompress coords in " + hpcbed)
    commandwords = ['liftOver', hpcbed, chainfile, decompbed, decompunmapped]
    subprocess.run(commandwords)

def write_gfa_fasta(nodegfa:str, nodefasta:str):

    with open(nodefasta, "w") as nfh:
        with open(nodegfa, "r") as gfh:
            for line in gfh:
                fields = line.split()
                #print(fields[0])
                if fields[0]=="S":
                    nodename = fields[1]
                    nodeseq = fields[2]
                    nfh.write(">" + nodename + "\n" + nodeseq + "\n")

    return(0)




import sys
import os
import subprocess
import re
import logging

logger = logging.getLogger(__name__)

def align_assembly_to_benchmark_haplotypes(queryfasta:str, outputfiles:dict, benchparams, args):
    matbenchfasta = benchparams["matbenchmark"]
    patbenchfasta = benchparams["patbenchmark"]
    if (args.aligner == "minimap2"):
        if not os.path.exists(outputfiles["aligntomatbenchprefix"] + ".mm2defparams.sort.bam"):
            minimap2_align(queryfasta, matbenchfasta, outputfiles["aligntomatbenchprefix"], args)
        if not os.path.exists(outputfiles["aligntopatbenchprefix"] + ".mm2defparams.sort.bam"):
            minimap2_align(queryfasta, patbenchfasta, outputfiles["aligntopatbenchprefix"], args)

        matbamfile = outputfiles["aligntomatbenchprefix"] + ".mm2defparams.sort.bam"
        patbamfile = outputfiles["aligntopatbenchprefix"] + ".mm2defparams.sort.bam"
    elif (args.aligner == "winnowmap2" or args.aligner == "winnowmap"):
        if not os.path.exists(outputfiles["aligntomatbenchprefix"] + ".wm2defparams.sort.bam"):
            winnowmap2_align(queryfasta, matbenchfasta, outputfiles["aligntomatbenchprefix"], benchparams["winnowmaprepkmers"], args)
        if not os.path.exists(outputfiles["aligntopatbenchprefix"] + ".wm2defparams.sort.bam"):
            winnowmap2_align(queryfasta, patbenchfasta, outputfiles["aligntopatbenchprefix"], benchparams["winnowmaprepkmers"], args)

        matbamfile = outputfiles["aligntomatbenchprefix"] + ".wm2defparams.sort.bam"
        patbamfile = outputfiles["aligntopatbenchprefix"] + ".wm2defparams.sort.bam"
    elif (args.aligner == "wfmash"):
        if not os.path.exists(outputfiles["aligntomatbenchprefix"] + ".wfmdefparams.sort.bam"):
            wfmash_align(queryfasta, matbenchfasta, outputfiles["aligntomatbenchprefix"], args)
        if not os.path.exists(outputfiles["aligntopatbenchprefix"] + ".wfmdefparams.sort.bam"):
            wfmash_align(queryfasta, patbenchfasta, outputfiles["aligntopatbenchprefix"], args)

        matbamfile = outputfiles["aligntomatbenchprefix"] + ".wfmdefparams.sort.bam"
        patbamfile = outputfiles["aligntopatbenchprefix"] + ".wfmdefparams.sort.bam"
    else:
        logger.critical("Unrecognized aligner " + args.aligner)
        print("Unrecognized aligner " + args.aligner)
        exit(1)

    return [matbamfile, patbamfile]

def align_haplotype_to_haplotype(queryfasta:str, reffasta:str, outputprefix:str, benchparams, args):
    if (args.aligner == "minimap2"):
        outputbam = outputprefix + ".mm2defparams.sort.bam"
        if not os.path.exists(outputbam):
            logger.info("Output bam file " + outputbam + " does not exist--aligning with " + args.aligner)
            minimap2_align(queryfasta, reffasta, outputprefix, args)
        else:
            logger.info("Skipping alignment of " + queryfasta + " to " + reffasta + " into bam file " + outputbam + " because it already exists!")
    elif (args.aligner == "winnowmap2" or args.aligner == "winnowmap"):
        outputbam = outputprefix + ".wm2defparams.sort.bam"
        if not os.path.exists(outputbam):
            winnowmap2_align(queryfasta, reffasta, outputprefix, benchparams["winnowmaprepkmers"], args)
        else:
            logger.info("Skipping alignment of " + queryfasta + " to " + reffasta + " into bam file " + outputbam + " because it already exists!")
    elif (args.aligner == "wfmash"):
        outputbam = outputprefix + ".wfmdefparams.sort.bam"
        if not os.path.exists(outputbam):
            wfmash_align(queryfasta, reffasta, outputprefix, args)
        else:
            logger.info("Skipping alignment of " + queryfasta + " to " + reffasta + " into bam file " + outputbam + " because it already exists!")
    else:
        logger.critical("Unrecognized aligner " + args.aligner)
        print("Unrecognized aligner " + args.aligner)
        exit(1)

    return outputbam

def minimap2_align(queryfasta:str, benchfasta:str, prefix:str, args)->list:
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    currentdir = os.getcwd()
    command = "minimap2 -a -t" + str(args.t) + " -x asm5 " + benchfasta + " " + queryfasta + " | samtools view -O BAM | samtools sort --threads " + str(args.t) + " -O bam -o " + prefix + ".mm2defparams.sort.bam > " + prefix + ".minimap2.defparams.out 2>&1"
    print("Running " + command)
    logger.debug("Running " + command)
    proc = subprocess.Popen(command, shell=True, env=env)
    proc.wait()

    index_bam_file(prefix + ".mm2defparams.sort.bam")

    return 0

def wfmash_align(queryfasta:str, benchfasta:str, prefix:str, args)->list:
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    currentdir = os.getcwd()
    command = "wfmash -t" + str(args.t) + " -Y \'#\' --sam " + benchfasta + " " + queryfasta + " | samtools view -O BAM | samtools sort --threads " + str(args.t) + " -O bam -o " + prefix + ".wfmdefparams.sort.bam > " + prefix + ".wfmdefparams.out 2>&1"
    print("Running " + command)
    logger.debug("Running " + command)
    proc = subprocess.Popen(command, shell=True, env=env)
    proc.wait()

    index_bam_file(prefix + ".wfmdefparams.sort.bam")

    return 0

def winnowmap2_align(queryfasta:str, benchfasta:str, prefix:str, repk19file:str, args)->list:
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    currentdir = os.getcwd()
    command = "winnowmap -W " + repk19file + " -a -t" + str(args.t) + " -I12g -x asm5 " + benchfasta + " " + queryfasta + " | samtools view -O BAM | samtools sort --threads " + str(args.t) + " -O bam -o " + prefix + ".wm2defparams.sort.bam > " + prefix + ".winnowmap2.defparams.out 2>&1"
    print("Running " + command)
    logger.debug("Running " + command)
    proc = subprocess.Popen(command, shell=True, env=env)
    proc.wait()

    index_bam_file(prefix + ".wm2defparams.sort.bam")

    return 0

def index_bam_file(bamfile:str):
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    currentdir = os.getcwd()
    command = "samtools index " + bamfile
    print("Running " + command)
    logger.debug("Running " + command)
    proc = subprocess.Popen(command, shell=True, env=env)
    proc.wait()




import sys
import os
import pysam
import argparse
import pybedtools
import logging
from pathlib import Path
from collections import namedtuple
from GQC import alignparse
from GQC import errors

logger = logging.getLogger(__name__)

# create namedtuple for bed intervals:
varianttuple = namedtuple('varianttuple', ['chrom', 'start', 'end', 'name', 'vartype', 'excluded', 'qvscore']) 

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="For a bam file, report all within-alignment discrepancies (single base mismatches, insertions, and deletions"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('--bam', required=True, help='bam file of alignments of sequences mapped to the reference fasta specified with --ref')
    parser.add_argument('--ref', type=str, required=True, help='(indexed) reference fasta file for the bam file reference')
    parser.add_argument('--query', type=str, required=True, help='(indexed) query fasta file that was aligned to the bam file reference')
    parser.add_argument('-m', '--minalignlength', type=int, required=False, default=5000, help='minimum length of alignment required to be included in het site gathering')
    parser.add_argument('-p', '--prefix', type=str, required=False, default="bamdiscrepancies", help='prefix to use in output filenames')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging purposes')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def read_aligns(bamobj, args):
    align_coords = {}

    for align in bamobj.fetch():
        if align.is_secondary:
            continue
        if align.reference_length >= args.minalignlength:
            query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
            querycoords = query + ":" + str(querystart) + "-" + str(queryend)
            refcoords = ref + ":" + str(refstart) + "-" + str(refend)
            querychrom = query.split("_")[0]
            refchrom = ref.split("_")[0]

            if querychrom == refchrom:
                align_coords[querycoords] = refcoords

    return align_coords

def main() -> None:

    args = parse_arguments(sys.argv[1:])

    logfile = args.prefix + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output to ' + logfile + ' for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    # read aligns from the BAM file
    alignobj = pysam.AlignmentFile(args.bam, "rb")

    # reference
    refobj = pysam.FastaFile(args.ref)
    queryobj = pysam.FastaFile(args.query)

    alignmapping = {}

    bedoutput = args.prefix + ".refcovered.bed"
    vcfoutput = args.prefix + ".discrepancies.vcf"

    with open(vcfoutput, "w") as vfh, open(bedoutput, "w") as bfh:
        for align in alignobj.fetch():
            if align.is_secondary:
                continue
            else:
                print("Primary alignment")

            query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
            refcoords = ref + ":" + str(refstart) + "-" + str(refend)
            querycoords = query + ":" + str(querystart) + "-" + str(queryend)
            if strand == 'F':
                bedstrand = '+'
            else:
                bedstrand = '-'

            mapqual = align.mapq

            bfh.write(ref + "\t" + str(refstart-1) + "\t" + str(refend) + "\t" + querycoords + "_" + refcoords + "_" + strand + "\t" + str(mapqual) + "\t" + bedstrand + "\n")

            logger.info("Parsing alignment between " + querycoords + " and " + refcoords)
            allvars = alignparse.align_variants(align, queryobj, query, querystart, queryend, refobj, ref, refstart, refend, strand)
            numvars = len(allvars)
            logger.info("Found " + str(numvars) + " variants in align with ref " + refcoords + " and query " + querycoords)
            for variant in allvars:
                vcfrecord = errors.vcf_format(variant, refobj, queryobj)
                vfh.write(vcfrecord)


if __name__ == "__main__":
    main()



import os
import re
import pybedtools
import pysam
import logging

logger = logging.getLogger(__name__)

def mergebed(bedfile:str, collapsecolumns='4', collapseoutput='collapse', collapsedelim='|', skipifexists=True)->str:

    mergedbed = bedfile.replace(".bed", ".merged.bed")

    if not skipifexists or not os.path.exists(mergedbed) or os.path.getsize(mergedbed) == 0:
        logger.debug("Merging " + bedfile + " to create " + mergedbed)

        unmergedints = pybedtools.bedtool.BedTool(bedfile)
        numunmerged = unmergedints.count()
   
        logger.debug("There are " + str(numunmerged) + " unmerged intervals")

        if numunmerged > 0:
            mergedints = unmergedints.merge(c=collapsecolumns, o=collapseoutput, delim=collapsedelim)
            mergedints.saveas(mergedbed)
        else:
            with open(mergedbed, 'a'):
                os.utime(mergedbed, None) 
            mergedints = unmergedints
    else:
        logger.debug("Skipping merging of " + bedfile + ": merged bedfile " + mergedbed + " already exists")
        mergedints = pybedtools.bedtool.BedTool(mergedbed)

    return [mergedints, mergedbed]

def mergeintervals(intervals, collapsecolumns='4', collapseoutput='collapse', collapsedelim='|', distance=0):

    mergedints = intervals.merge(c=collapsecolumns, o=collapseoutput, delim=collapsedelim, d=distance)

    return mergedints

def mergemultiplebedfiles(bedfilelist:list, sort=True, postmerge=False):

    if len(bedfilelist) < 2:
        logger.critical("Cannot call mergemultiplebedfiles on less than two bed files!")
        exit(1)
    #if len(bedfilelist) > 2:
        #logger.critical("Theres a bug in GQC bedtoolslib.py so it cant yet merge multiple excluded regions bed files!")
        #exit(1)

    #firstbedtool = pybedtools.bedtool.BedTool(bedfilelist[0])
    #allbedtools = firstbedtool.cat(bedfilelist[1])
    allbedtools = pybedtools.bedtool.BedTool(bedfilelist[0])
    for i in range(len(bedfilelist) - 1):
        allbedtools = allbedtools.cat(bedfilelist[i+1])

    if sort:
        return allbedtools.sort()
    else:
        return allbedtools

def bedsum(intervals)->int:

    alllengths = map(len, intervals)
    bedsum = sum(alllengths)

    return bedsum

def intersectbed(bedfile1:str, bedfile2:str, outputfile:str, writefirst=False, writeboth=False, outerjoin=False, requirewhole=False, v=False)->str:

    logger.debug("Intersecting " + bedfile1 + " and " + bedfile2 + " to create " + outputfile)

    command = "bedtools intersect -a " + bedfile1 + " -b " + bedfile2
    if writefirst:
        command = command + " -wa"
    if writeboth:
        command = command + " -wo"
    if v:
        command = command + " -v"
    if outerjoin:
        command = command + " -loj"
    if requirewhole:
        command = command + " -f 1.0"
    os.system(command + " > " + outputfile)
    intersectbed = pybedtools.bedtool.BedTool(outputfile).sort()
    intersectbed.saveas(outputfile)

    return [intersectbed, outputfile]

def intersectintervals(intervals1, intervals2, v=False, wo=False, wa=False, wb=False, counts=False):

    if wo:
        print("Intersect called with -wo option")
    intersectedints = intervals1.intersect(intervals2, v=v, wa=wa, wb=wb, wo=wo, c=counts)

    return intersectedints

def subtractintervals(intervals1, intervals2, A=False, wb=False, wo=False):

    subtractedints = intervals1.subtract(intervals2, A=A, wb=wb, wo=wo)

    return subtractedints

def genomeintervals(fastafile:str):

    indexfile = fastafile + ".fai"
    if not os.path.exists(indexfile):
        pysam.faidx(fastafile)

    fastaobj = pysam.FastaFile(fastafile)
    bedstring = ""
    for refname in fastaobj.references:
        reflength = fastaobj.get_reference_length(refname)
        bedstring = bedstring + refname + "\t0\t" + str(reflength) + "\n"
    
    return pybedtools.BedTool(bedstring, from_string=True)

def binintervals(intervals, binsize=10000, outbedfile=None):
    binbedtring = ""
    for interval in intervals:
        chrom = interval.chrom
        start = interval.start
        end = interval.end

        binstart = start
        binend = start + binsize
        while binend <= end:
            binbedstring = binbedstring + chrom + "\t" + str(binstart) + "\t" + str(binend) + "\n"
            binstart = binend
            binend = binend + binsize

    binnedbedobj = pybedtools.BedTool(binbedstring, from_string = True)
    if outbedfile is not None:
        binnedbedobj.sort().saveas(outbedfile)

    return binnedbedobj



import sys
import os
import re
import shutil
import pysam
import argparse
import pybedtools
import importlib.resources
from pathlib import Path
from GQC import alignparse

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Examine bam files aligning one diploid haplotype to another and report on synteny/heterozygosity"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('--bam1', required=True, help='bam file of alignments in one direction from one haplotype mapped to the other')
    parser.add_argument('--bam2', required=True, help='bam file of alignments in the other direction')
    parser.add_argument('--ref1', type=str, required=True, help='(indexed) reference fasta file for the first bam file')
    parser.add_argument('--ref2', type=str, required=True, help='(indexed) reference fasta file for the first bam file')
    parser.add_argument('-m', '--minalignlength', type=int, required=False, default=5000, help='minimum length of alignment required to be included in het site gathering')
    parser.add_argument('--hetwindowsize', type=int, required=False, default=20000, help='window size for reporting heterozygosity levels')
    parser.add_argument('-p', '--prefix', type=str, required=False, default="hetsites", help='prefix to use in output filenames')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def read_aligns(bamobj, args):
    align_coords = {}

    for align in bamobj.fetch():
        if align.is_secondary:
            continue
        if align.reference_length >= args.minalignlength:
            query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
            querycoords = query + ":" + str(querystart) + "-" + str(queryend)
            refcoords = ref + ":" + str(refstart) + "-" + str(refend)
            querychrom = query.split("_")[0]
            refchrom = ref.split("_")[0]

            if querychrom == refchrom:
                align_coords[querycoords] = refcoords

    return align_coords

def find_hets_and_homregions(bamobj, refobj, queryobj, alignmap:dict, args):

    # strings which will be used to create bed files with pybedtools:
    aligncoveredregions = ""
    aligncoveredwindows = ""
    hetvariants = ""
    homregions = ""
    for align in bamobj.fetch():
        query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
        querycoords = query + ":" + str(querystart) + "-" + str(queryend)
        refcoords = ref + ":" + str(refstart) + "-" + str(refend)

        if querycoords in alignmap.keys() and alignmap[querycoords] == refcoords:
            # add query\tstart\tstop\tname\tscore\tstrand to covered string:
            coveredstring = ref + "\t" + str(refstart - 1) + "\t" + str(refend) + "\t" + querycoords + "/" + strand + "\t0.0\t" + strand + "\n"
            aligncoveredregions = aligncoveredregions + coveredstring
            coveredwindows = make_covered_window_string(ref, refstart, refend, query, querystart, queryend, strand, args)
            aligncoveredwindows = aligncoveredwindows + coveredwindows
            variantlist = alignparse.align_variants(align, queryobj, query, querystart, queryend, refobj, ref, refstart, refend, strand)
            lastquerypos = querystart
            for variant in variantlist:
                chrom = variant.chrom
                start = variant.start
                end = variant.end
                name = variant.name
                variantstring = chrom + "\t" + str(start) + "\t" + str(end) + "\t" + name + "\n"
                hetvariants = hetvariants + variantstring

    return aligncoveredregions, aligncoveredwindows, hetvariants, homregions

def make_covered_window_string(ref:str, refstart:int, refend:int, query:str, querystart:int, queryend:int, strand:str, args):
    bedstring = ""
    windowsize = args.hetwindowsize

    windowstart = windowsize*int(refstart/windowsize + 1.0)
    lastwindowend = windowsize*int(refend/windowsize)
    windownumber = 1
    while windowstart < lastwindowend:
        windowname = ref + "_" + str(refstart) + "_" + str(refend) + "_" + query + "_" + str(querystart) + "_" + str(queryend) + "_" + strand + "_" + str(windownumber)
        bedstring = bedstring + ref + "\t" + str(windowstart) + "\t" + str(windowstart + windowsize) + "\t" + windowname + "\n"
        windownumber = windownumber + 1
        windowstart = windowstart + windowsize
    return bedstring

def main() -> None:

    args = parse_arguments(sys.argv[1:])

    # read aligns from BAM files
    # coords dictionaries have keys that are the query coordinates and values that are the ref coordinates
    alignobj1 = pysam.AlignmentFile(args.bam1, "rb")
    coords1 = read_aligns(alignobj1, args)
    alignobj2 = pysam.AlignmentFile(args.bam2, "rb")
    coords2 = read_aligns(alignobj2, args)
    # refs
    refobj1 = pysam.FastaFile(args.ref1)
    refobj2 = pysam.FastaFile(args.ref2)

    alignmapping = {}
    for align1 in alignobj1.fetch():
        ref1, refstart1, refend1, query1, querystart1, queryend1, strand1 = alignparse.retrieve_align_data(align1)
        refcoords1 = ref1 + ":" + str(refstart1) + "-" + str(refend1)
        querycoords1 = query1 + ":" + str(querystart1) + "-" + str(queryend1)

        if querycoords1 in coords2.keys():
            refcoords2 = coords2[querycoords1]
            if refcoords1 == refcoords2 and refend1 - refstart1 >= 100000:
                alignmapping[refcoords1] = querycoords1
                alignmapping[querycoords1] = refcoords1

    coveredregions1, coveredwindows1, hetvariants1, homregions1 = find_hets_and_homregions(alignobj1, refobj1, refobj2, alignmapping, args)
    coveredregions2, coveredwindows2, hetvariants2, homregions2 = find_hets_and_homregions(alignobj2, refobj2, refobj1, alignmapping, args)

    allvars = hetvariants1 + hetvariants2
    variantintervals = pybedtools.BedTool(allvars, from_string=True)
    variantintervals.sort().saveas(args.prefix + ".hetsites.bed")
    allcovered = coveredregions1 + coveredregions2
    coveredintervals = pybedtools.BedTool(allcovered, from_string=True)
    coveredintervals.sort().saveas(args.prefix + ".1to1alignable.bed")
    allcoveredwindows = coveredwindows1 + coveredwindows2
    coveredwindowintervals = pybedtools.BedTool(allcoveredwindows, from_string=True)
    coveredwindowintervals.sort().saveas(args.prefix + ".1to1alignable.windows.bed")

if __name__ == "__main__":
    main()



import sys
import importlib.util
import pybedtools
import logging

if (spec := importlib.util.find_spec('matplotlib')) is not None:
    #importing the present module
    module = importlib.util.module_from_spec(spec)
    sys.modules['matplotlib'] = module
    spec.loader.exec_module(module)
    matlibimported = True
else:
    print(f"To print heatmap colors in bedfiles, you need to install matplotlib")
    matlibimported = False

if matlibimported:
   import matplotlib.pyplot as plt
   import matplotlib.colors as mcolors

logger = logging.getLogger(__name__)

def add_colors_to_intervals(windowedintervalswithcounts, args):
    if not matlibimported:
        return None

    intervalbedstringwithcolors = ""

    allscores = []
    for interval in windowedintervalswithcounts:
        score = int(interval[4])
        if args.windowsize > interval.end - interval.start:
            score = int(score * args.windowsize / (interval.end - interval.start))
        allscores.append(score)

    if args.heatmapmincount and args.heatmapmaxcount:
        minscore = args.heatmapmincount
        maxscore = args.heatmapmaxcount
    else:
        numintervals = len(allscores)
        lowcutoff = int(0.05*numintervals)
        highcutoff = int(0.95*numintervals)
        allscores.sort()
        minscore = allscores[lowcutoff]
        maxscore = allscores[highcutoff]
    logger.info("Using count " + str(minscore) + " for minimum end of spectrum and count " + str(maxscore) + " for maximum end of spectrum")

    for interval in windowedintervalswithcounts:
        ref = interval.chrom
        start = interval.start
        end = interval.end
        name = interval.name
        score = int(interval[4])

        rgbval = convert_score_to_rgb(score, min_score=minscore, max_score=maxscore)

        intervalbedstringwithcolors = intervalbedstringwithcolors + ref + "\t" + str(start) + "\t" + str(end) + "\t" + name + "\t" + str(score) + "\t.\t" + str(start) + "\t" + str(end) + "\t" + rgbval + "\n"

    intervalbedtool =  pybedtools.BedTool(intervalbedstringwithcolors, from_string=True)

    return intervalbedtool

# Function to convert a score to an RGB color (adapted from a function written by Steven Solar)
def convert_score_to_rgb(score:int, min_score=0, max_score=1000, colormap='YlOrRd')->str:
    if score < min_score:
        score = min_score
    elif score > max_score:
        score = max_score

    # Define the colormap
    cmap = plt.get_cmap(colormap)

    # Define the normalization range
    norm = mcolors.Normalize(vmin=min_score, vmax=max_score)

    # Get the RGBA color corresponding to the percent identity
    rgba = cmap(norm(score))

    # Convert RGBA to RGB
    rgb = tuple(int(255 * x) for x in rgba[:3])
    return str(rgb[0]) + "," + str(rgb[1]) + "," + str(rgb[2])





import sys
import os
import re
import shutil
import pysam
import argparse
import logging
from pathlib import Path
from GQC import output
from GQC import variants

logger = logging.getLogger(__name__)

def check_for_executable(programname:str):
    if shutil.which(programname) is None:
        print("You don\'t seem to have " + programname + " in your path. Please install " + programname)
        logger.critical("You don\'t seem to have " + programname + " in your path. Please install " + programname)
        exit(1)
    return 0

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Construct a transformed genome fasta from a VCF file, masking regions not included in a high-confidence BED file"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('-g', '--gvcf', required=False, default=None, help='gVCF file from which VCF variants and high confidence regions will be extracted')
    parser.add_argument('--mingq', type=int, required=False, default=10, help='if parsing a gVCF file, min GQ value to require to include variants and/or reference regions')
    parser.add_argument('--vcf', required=False, default=None, help='VCF file of variants to be applied to a reference when constructing the transformed genome')
    parser.add_argument('--hcregions', type=str, required=False, default=None, help='bed file of high confidence regions (in coordinates of the VCF reference)')
    parser.add_argument('-r', '--reffasta', type=str, required=True, help='(indexed) fasta file for the VCF file reference genome')
    parser.add_argument('--hap1chroms', type=str, required=False, default='chrY', help='comma-delimited list of reference entries that are only on haplotype 1')
    parser.add_argument('--hap2chroms', type=str, required=False, default='chrX', help='comma-delimited list of reference entries that are only on haplotype 2')
    parser.add_argument('-s', '--samplename', type=str, required=True, help='name of sample in VCF file')
    parser.add_argument('-p', '--prefix', type=str, required=True, help='prefix for output directory name')
    parser.add_argument('-t', type=int, required=False, default=2, help='number of processors to use when running multi-threaded programs')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging purposes')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def main() -> None:

    args = parse_arguments(sys.argv[1:])

    logfile = args.prefix + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    # check for necessary installed programs and write an output directory:
    check_for_executable("bedtools")
    check_for_executable("bcftools")
    check_for_executable("liftOver")

    # did the user pass a single gVCF file or a VCF/high conf bed combo?
    if args.gvcf:
        gvcffile = args.gvcf
        logger.info("Creating VCF and high-confidence BED file from gVCF file " + gvcffile)
        # here we'll call a new routine to do this
        [vcffile, hcbedfile] = variants.create_vcf_and_hcbed_from_gvcf(gvcffile, args)
    elif args.vcf and args.hcregions:
        vcffile = Path(args.vcf)
        hcbedfile = Path(args.hcregions)
        if not vcffile.is_file():
            logger.critical("VCF file " + args.vcffile + " must exist and be readable")
            print("VCF file " + args.vcffile + " must exist and be readable", file=sys.stderr)
            exit(1)
    else:
        logger.critical("Either a gvcf file (--gvcf option) or both a VCF file and a high-confidence BED file (options --vcf, --hcregions) must be specified")
        print("VCF file " + args.vcffile + " must exist and be readable", file=sys.stderr)
        exit(1)


    # pysam objects for the benchmark and test assembly fasta files:
    ref = Path(args.reffasta)
    if not ref.is_file():
        logger.critical("Ref fasta file " + args.reffasta + " must exist and be readable")
        print("Ref fasta file " + args.reffasta + " must exist and be readable", file=sys.stderr)
        exit(1)
    outputdir = output.create_output_directory(args.prefix)

    # dictionary of this run's output file names:
    outputfiles = output.name_constructgenome_output_files(args, outputdir)

    # apply VCF variants to the reference, then filter non-high-confidence regions with Ns:
    variants.create_constructed_genome(args.reffasta, args.vcf, args.hcregions, outputdir, outputfiles, args)

if __name__ == "__main__":
    main()



import sys
import os
import shutil
import re
import argparse
import logging
from GQC import kmers

logger = logging.getLogger(__name__)

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Find kmers with extreme ATGC composition in a FastK database"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('--fastkdb', required=True, help='FastK database root')
    parser.add_argument('-p', '--prefix', type=str, required=False, default="hetsites", help='prefix to use in output filenames')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging purposes')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def check_for_histex():
    if shutil.which("Tabex") is None:
        print("You don\'t seem to have Tabex in your path. Please install Tabex")
        logger.critical("You don\'t seem to have Tabex in your path. Please install Tabex")
        exit(1)
    return 0

def find_extreme_kmers(fastkdbroot:str, args):

    kmers.find_extreme_kmers(fastkdbroot)

    return 0

def main() -> None:

    args = parse_arguments(sys.argv[1:])

    logfile = args.prefix + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    check_for_histex()

    find_extreme_kmers(args.fastkdb, args)

if __name__ == "__main__":
    main()





import sys
import logging

logger = logging.getLogger(__name__)

# reminders: in aligndata, (1) all coordinates are 1-based, (2) strand is "+" or "-", (3) querystart is the query's lower coordinate,
# so doesn't correspond to targetstart if alignment is on the reverse strand

def write_structural_errors(aligndata:list, refobj, queryobj, outputdict, bmstats, args)->str:

    aligndict = {}
    current_align = None
    with open(outputdict["structvariantbed"], "w") as sfh:
        for align in sorted(aligndata, key=lambda a: (a["target"], a["targetstart"], a["targetend"])):
            refentry = align["target"]
            if refentry not in aligndict:
                aligndict[refentry] = [align]
            else:
                aligndict[refentry].append(align)
            query = align["query"]
            refstart = align["targetstart"]
            refend = align["targetend"]
            querystart = align["querystart"]
            queryend = align["queryend"]
            strand = align["strand"]
            if current_align is not None:
                refdiff = refstart - current_align["targetend"]
                if refentry == current_align["target"] and query == current_align["query"] and strand == current_align["strand"]:
                    if strand == "+":
                        querydiff = querystart - current_align["queryend"]
                        query1 = current_align["queryend"]
                        query2 = querystart
                    else:
                        querydiff = queryend - current_align["querystart"]
                        query1 = querystart
                        query2 = current_align["queryend"]
   
                    netdiff = querydiff - refdiff
                    if refdiff < querydiff: # refdiff less than querydiff (insertion), netshift positive
                        if refdiff > 0:
                            sfh.write(refentry + "\t" + str(current_align["targetend"] - 1) + "\t" + str(refstart) + "\tSameContigInsertion\t" + query + "\t" + str(query1) + "\t" + str(query2) + "\t" + str(current_align["targetend"]) + "\t" + str(refstart) + "\t" + str(netdiff) + "\t" + strand + "\n")
                        else:
                            sfh.write(refentry + "\t" + str(refstart - 1) + "\t" + str(current_align["targetend"]) + "\tSameContigInsertion\t" + query + "\t" + str(query1) + "\t" + str(query2) + "\t" + str(current_align["targetend"]) + "\t" + str(refstart) + "\t" + str(netdiff) + "\t" + strand + "\n")
                    else: # refdiff greater than than querydiff (deletion), netshift negative
                        if refdiff > 0:
                            sfh.write(refentry + "\t" + str(current_align["targetend"] - 1) + "\t" + str(refstart) + "\tSameContigDeletion\t" + query + "\t" + str(query1) + "\t" + str(query2) + "\t" + str(current_align["targetend"]) + "\t" + str(refstart) + "\t" + str(netdiff) + "\t" + strand + "\n")
                        else:
                            sfh.write(refentry + "\t" + str(refstart - 1) + "\t" + str(current_align["targetend"]) + "\tSameContigDeletion\t" + query + "\t" + str(query1) + "\t" + str(query2) + "\t" + str(current_align["targetend"]) + "\t" + str(refstart) + "\t" + str(netdiff) + "\t" + strand + "\n")
    
                elif refentry == current_align["target"]: # strand switch or new contig:
                    queryentries = query + "/" + current_align["query"]
                    strands = strand + "/" + current_align["strand"]
                    if refdiff > 0:
                        sfh.write(refentry + "\t" + str(current_align["targetend"] - 1) + "\t" + str(refstart) + "\tBetweenContigDeletion\t" + queryentries + "\t.\t.\t" + str(current_align["targetend"]) + "\t" + str(refstart) + "\tNA\t" + strands + "\n")
                    else:
                        sfh.write(refentry + "\t" + str(refstart - 1) + "\t" + str(current_align["targetend"]) + "\tBetweenContigInsertion\t" + queryentries + "\t.\t.\t" + str(current_align["targetend"]) + "\t" + str(refstart) + "\tNA\t" +strands + "\n")
            current_align = align

    return 0




import os
import sys
import re
import subprocess
import logging
from pathlib import Path

logger = logging.getLogger(__name__)

def create_kmer_database(fastafile:str, outputdir:str, prefix:str, kmersize=40):
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    kmerdbroot = outputdir + "/" + prefix + ".kmers.k" + str(kmersize)
    if not os.path.exists(kmerdbroot + ".ktab"):
       command = "FastK -k" + str(kmersize) + " -T2 -N" + kmerdbroot + " -p -t1 -v " + fastafile
       print("Running: " + command)
       logger.info("Running: " + command)
       proc = subprocess.Popen(command, shell=True, env=env)
       proc.wait()

def remove_kmer_database(outputdir:str, prefix:str, kmersize=40):
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    kmerdbroot = outputdir + "/" + prefix + ".kmers.k" + str(kmersize)
    if os.path.exists(kmerdbroot + ".ktab"):
       command = "Fastrm " + kmerdbroot
       print("Running: " + command)
       logger.info("Running: " + command)
       proc = subprocess.Popen(command, shell=True, env=env)
       proc.wait()

def find_anotb_kmers(db1prefix:str, db2prefix:str, outputdir:str, prefix:str, kmersize=40):
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    kmerdbroot = outputdir + "/" + prefix + ".kmers.k" + str(kmersize)
    if not os.path.exists(kmerdbroot + ".ktab"):
       command = "Logex -T2 -h \"" + kmerdbroot + " = A-B\" " + outputdir + "/" + db1prefix + ".kmers.k" + str(kmersize) + " " + outputdir + "/" + db2prefix + ".kmers.k" + str(kmersize)
       print("Running: " + command)
       logger.info("Running: " + command)
       proc = subprocess.Popen(command, shell=True, env=env)
       proc.wait()

    return kmerdbroot

def map_kmer_markers_onto_fasta(fastafile:str, markerfilelist:list, outputdir:str):
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()

    outputbedlist = []
    tmpdir = outputdir + "/tmp"
    path = Path(tmpdir)
    logger.info("Creating temporary directory " + tmpdir + " for output")
    path.mkdir(exist_ok=True)

    pattern = r"\.fastq$|\.fasta$|\.fq$|\.fa$|\.fastq.gz$|\.fasta.gz$|\.fq.gz$|\.fa.gz$"
    assemblyroot = re.sub(pattern, "", fastafile)
    outputfile = outputdir + "/" + assemblyroot + ".kmers.merge.bed"

    if not os.path.exists(outputfile):
        mergefilestring = ""
        for markerfile in markerfilelist:
            # KmerMap constructs the output file name from the third argument with ".assemblyroot" + ".kmers.merge.bed"
            kmermapinputprefix = markerfile
            kmermapoutputprefix = markerfile + "." + assemblyroot
            mergefilestring = mergefilestring + " " + kmermapoutputprefix + ".kmers.merge.bed"
            markercommand = "KmerMap -v -m -T2 -P" + tmpdir + " " + markerfile + " " + fastafile + " " + kmermapinputprefix
            
            if not os.path.exists(kmermapoutputprefix + ".kmers.merge.bed"):
                logger.info("Running: " + markercommand)
                proc = subprocess.Popen(markercommand, shell=True, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = proc.communicate()
                proc.wait()
    
                if proc.returncode == 0:
                    logger.info(markercommand + " completed successfully")
                else:
                    logger.info(stderr.decode())
            else:
                logger.info("Skipping run of " + markercommand + ": output already exists")
    
        logger.debug("Merging output files into " + outputfile)
        mergesortcommand = "cat" + mergefilestring + " | sort -k1,1n -k2,2n -k3,3n " + " > " + outputfile
        logger.debug(mergesortcommand)
        os.system(mergesortcommand)
        cleanupcommand = "rm" + mergefilestring
        logger.debug(cleanupcommand)
        os.system(cleanupcommand)
    else:
        logger.debug("Skipping kmer mapping and merging of kmer files into " + outputfile + ": output already exists")
   
    return outputfile

def find_extreme_kmers(fastkdbroot:str):
    if os.path.exists(fastkdbroot + ".hist"):
        env = os.environ.copy()
        env['LD_LIBRARY_PATH'] = os.getcwd()
        command = "Tabex -t2 " + fastkdbroot + " LIST"
        print("Running: " + command)
        logger.info("Running: " + command)
        proc = subprocess.Popen(command, shell=True, env=env)
        while True:
            line = proc.stdout.readline()
            if not line:
                break
            print(line.strip())
        proc.wait()
    else:
        logger.critical("No such fastk database: " + fastkdbroot)

# returns the total count of base1 and base2 on the two strands of a kmer string
def twobase_kmer_comp(base1:str, base2:str, kmerseq:str):
    kmerseqlc = kmerseq.tolower()
    base1 = base1.tolower()
    base2 = base2.tolower()
    compbases = {'a':'t', 't':'a', 'g':'c', 'c':'g'}

    fwdcount = kmerseq.count(base1) + kmerseq.count(base2)
    revcount = kmerseq.count(compbase[base1]) + kmerseq.count(compbase[base2])

    return [fwdcount, revcount]




import os
import re
import pysam
import pybedtools
import logging
from pathlib import Path
from GQC import bedtoolslib
from collections import namedtuple
import time

logger = logging.getLogger(__name__)

def write_genome_bedfiles(queryobj, refobj, args, benchparams, outputfiles, bedobjects):

    write_whole_genome_bedfile(queryobj, args, outputfiles, bedobjects, "testgenomeregions", "testgenomebed")
    write_whole_genome_bedfile(refobj, args, outputfiles, bedobjects, "benchgenomeregions", "benchgenomebed")
    write_excluded_bedfile(refobj, args, benchparams, outputfiles, bedobjects)
    # did the command-line arguments specify a pre-existing bedfile of N-stretch locations?
    if hasattr(args, 'n_bedfile'):
        user_n_file = args.n_bedfile
    else:
        user_n_file = None

    find_all_ns(queryobj, args, user_n_file, outputfiles, bedobjects, "testnregions", "testnonnregions", "testnbed", "testnonnbed")

    # if bench config file has an nstretchfile, use it and complement to find benchmark non-N files:
    find_all_ns(refobj, args, None, outputfiles, bedobjects, "benchnregions", "benchnonnregions", "benchnbed", "benchnonnbed")

# write_excluded_bedfile function is written for assembly benchmarking only, not read benchmarking:
def write_excluded_bedfile(refobj, args, benchparams, outputfiles, bedobjects):

    allexcludedbedfiles = []
    if args.excludefile is not None:
        allexcludedbedfiles.append(args.excludefile)
    if args.includefile is not None:
        write_nonincluded_file(refobj, args.includefile, outputfiles["nonincludedbed"], bedobjects)
        allexcludedbedfiles.append(outputfiles["nonincludedbed"])
    if "excluderegions" in benchparams.keys():
        configexcluderegions = benchparams["excluderegions"]
        configexcludepath = Path(configexcluderegions)
        if configexcludepath.is_file():
            allexcludedbedfiles.append(configexcluderegions)
        else:
            logger.warning("Exclude file " + configexcluderegions + " does not exist so will not be used")
    if "nstretchregions" in benchparams.keys():
        nexcluderegions = benchparams["nstretchregions"]
        nexcludepath = Path(nexcluderegions)
        if nexcludepath.is_file():
            allexcludedbedfiles.append(nexcluderegions)
        else:
            logger.warning("Benchmark N file " + nexcluderegions + " does not exist so will not be used")
    if len(allexcludedbedfiles) == 0:
        bedobjects["allexcludedregions"] = pybedtools.BedTool("", from_string = True)
    elif len(allexcludedbedfiles) == 1:
        bedobjects["allexcludedregions"] = pybedtools.BedTool(allexcludedbedfiles[0]).sort()
    elif len(allexcludedbedfiles) > 1:
        logger.info("Merging " + str(allexcludedbedfiles) + " to create a combined excluded regions file")
        bedobjects["allexcludedregions"] = bedtoolslib.mergemultiplebedfiles(allexcludedbedfiles)
    bedobjects["allexcludedregions"].saveas(outputfiles["allexcludedbed"])
    # now subtract the set of all excluded regions from the genome to obtain all regions included:
    bedobjects["allincludedregions"] = bedobjects["benchgenomeregions"].subtract(bedobjects["allexcludedregions"])
    bedobjects["allincludedregions"].saveas(outputfiles["includednonexcludedbed"])

    return 0

# write_included_bedfile function is written for read benchmarking only, not assembly benchmarking:
def write_included_bedfile(refobj, args, benchparams, outputfiles):

    if args.regions != "":
        benchintervals = pybedtools.BedTool(args.regions)
        logger.info("Restricting evaluation to read sequence aligning to regions in " + args.regions)
    else:
        genomebedstring = ""
        for chrom in refobj.references:
            chromlength = refobj.get_reference_length(chrom)
            genomebedstring = genomebedstring + chrom + "\t0\t" + str(chromlength) + "\n"
        benchintervals = pybedtools.BedTool(genomebedstring, from_string = True)

    if "excluderegions" in benchparams.keys():
        configexcluderegions = benchparams["excluderegions"]
        configexcludepath = Path(configexcluderegions)
        if configexcludepath.is_file():
            excludeintervals = pybedtools.BedTool(configexcludepath)
            logger.info("Not including read bases aligned to regions excluded in " + configexcluderegions)
            benchintervals = benchintervals.subtract(excludeintervals)

    benchintervals.saveas(outputfiles["includedbedfile"])
    logger.info("Benchmark regions included in this analysis are written to file " + outputfiles["includedbedfile"])

    return benchintervals

def write_whole_genome_bedfile(fastaobj, args, outputfiles, bedobjects, regionskey, regionfilekey):

    genomebedstring = ""
    for scaffold in fastaobj.references:
        scaffoldlength = fastaobj.get_reference_length(scaffold)
        scaffstring = scaffold + "\t0\t" + str(scaffoldlength) + "\n"
        genomebedstring += scaffstring
    bedobjects[regionskey] = pybedtools.BedTool(genomebedstring, from_string = True)

    # write bed file if it doesn't exist yet
    if not os.path.exists(outputfiles[regionfilekey]):
        bedobjects[regionskey].sort().saveas(outputfiles[regionfilekey])

    return 0

def write_nonincluded_file(refobj, includedbed, nonincludedbed, bedobjects):

    genomebedstring = ""
    for refentry in refobj.references:
        reflength = refobj.get_reference_length(refentry)
        genomebedstring = genomebedstring + refentry + "\t0\t" + str(reflength) + "\n"

    bedobjects["benchgenomeregions"] = pybedtools.BedTool(genomebedstring, from_string = True)
    bedobjects["includedregions"] = pybedtools.BedTool(includedbed)
    bedobjects["nonincludedregions"] = bedobjects["benchgenomeregions"].subtract(bedobjects["includedregions"])
    bedobjects["nonincludedregions"].saveas(nonincludedbed)

    return 0

# If existingnfile is passed in the third argument, it will be read and complemented to find non-N regions. If
#    no existing N file is passed, third argument must be "None"
def find_all_ns(fastaobj, args, user_n_file, outputfiles, bedobjects, nregionkey, nonnregionkey, nbedfilekey, nonnbedfilekey)->list:

    if not user_n_file:
        p = re.compile("N+")

    gapbedstring = ""
    contigbedstring = ""
    n_interval_dict = {}
    # if an n interval file has been specified in the command line options, read it into gapbedstring, and subtract from genome to get contigbedstring:
    if user_n_file:
        nbedobj = BedTool(user_n_file)
        for n_interval in nbedobj:
            chrom = n_interval.chrom
            start = n_interval.start
            end = n_interval.end
            gapname = n_interval.name
            gapbedstring += chrom + "\t" + str(start) + "\t" + str(end) + "\t" + gapname + "\n"
            if chrom in n_interval_dict.keys():
                n_interval_dict[chrom].append(n_interval)
            else:
                n_interval_dict[chrom] = [n_interval]
        for ref in fastaobj.references:
            if ref not in n_interval_dict.keys():
                n_interval_dict[ref] = []
            contignum = 1
            contigstart = 0
            for interval in n_interval_dict[ref]:
                interval_name = ref + "." + str(contignum)
                contigbedstring += ref + "\t" + str(contigstart) + "\t" + str(interval.start) + "\t" + interval_name + "\n"
                contignum = contignum + 1
                contigstart = interval.end
            refend = fastaobj.get_reference_length(ref)
            interval_name = ref + "." + str(contignum)
            contigbedstring += ref + "\t" + str(contigstart) + "\t" + str(refend) + "\t" + interval_name + "\n"
    else:
        for ref in fastaobj.references:
            contignum = 1
            contigstart = 0
            findstring = 'N' * args.minns
            chromseq = fastaobj.fetch(ref).upper()
            refend = fastaobj.get_reference_length(ref)
            start = chromseq.find(findstring)
            while start != -1:
                end = start + args.minns
                while end < refend and chromseq[end] == 'N':
                    end = end + 1
                if end - start >= args.minns:
                    gapname = "N." + ref + "." + str(contignum)
                    gapstring = ref + "\t" + str(start) + "\t" + str(end) + "\t" + gapname + "\n"
                    gapbedstring += gapstring
                    contigend = start
                    contigname = ref + "." + str(contignum)
                    contigbedstring += ref + "\t" + str(contigstart) + "\t" + str(contigend) + "\t" + contigname + "\n"
                    contignum = contignum + 1
                    contigstart = end
                start = chromseq.find(findstring, end, refend)

            contigname = ref + "." + str(contignum)
            contigbedstring += ref + "\t" + str(contigstart) + "\t" + str(refend) + "\t" + contigname + "\n"

        bedobjects[nonnregionkey] = pybedtools.BedTool(contigbedstring, from_string = True)
        bedobjects[nregionkey] = pybedtools.BedTool(gapbedstring, from_string = True)

    if outputfiles[nonnbedfilekey] and not os.path.exists(outputfiles[nonnbedfilekey]):
        bedobjects[nonnregionkey].sort().saveas(outputfiles[nonnbedfilekey])
    if outputfiles[nbedfilekey] and not os.path.exists(outputfiles[nbedfilekey]):
        bedobjects[nregionkey].sort().saveas(outputfiles[nbedfilekey])

    return 0

def revcomp(seq:str) -> str:
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', 'M': 'N', 'K': 'N', 'R': 'N', 'W': 'N', 'Y': 'N'}
    bases = list(seq)
    bases = bases[::-1]
    bases = [complement[base.upper()] for base in bases]

    return ''.join(bases)

def compress_sequence(fastafile:str)->str:
    refobj = pysam.FastaFile(fastafile)

    if args.prefix is not None:
        chainfile = args.prefix + ".hpc.chain"
        compressedfasta = args.prefix + ".hpc.fasta"
    else:
        chainfile = re.sub("\\.fa.*$", ".hpc.chain", fastafile)
        chainfile = re.sub(".*/", "", chainfile)
        compressedfasta = re.sub("\\.fa.*$", ".hpc.fasta", fastafile)
        compressedfasta = re.sub(".*/", "", compressedfasta)

    if os.path.isfile(chainfile) and os.path.isfile(compressedfasta):
        print("Using pre-existing chain file " + chainfile + " and compressed fasta file " + compressedfasta)
    else:
        print("Writing chain file and homopolymer compressed assembly fasta")

        cfh = open(chainfile, "w")
        with open(compressedfasta, "w") as hfh:
            refseqs = refobj.references
            repeatnum = 0
            chainid = 0
            for entry in refseqs:
                print("Processing " + entry)
                fullseq = refobj.fetch(entry).upper()
                reflength = len(fullseq)
                matches = 0
                insertions = 0
                lastbase = ""
                inhpc = False
                compressedstring = ""
                chains = []
                for base in fullseq:
                    if base != lastbase:
                        if insertions > 0:
                            chains.append(str(matches) + "\t0\t" + str(insertions) + "\n")
                            insertions = 0
                            matches = 0
                        matches = matches + 1
                        compressedstring = compressedstring + base
                        lastbase = base
                    else: # hp!
                        insertions = insertions + 1
                if matches > 0 and insertions == 0:
                    chains.append(str(matches) + "\n\n")
                elif matches > 0 and insertions > 0:
                    chains.append(str(matches) + "\t0\t" + str(insertions) + "\n0\n\n")

                chainid = chainid + 1
                hpclength = len(compressedstring)
                cfh.write("chain 4000 " + entry + ".compressed " + str(hpclength) + " + 0 " + str(hpclength) + " " + entry + " " + str(reflength) + " + 0 " + str(reflength) + " " + str(chainid) + "\n")
                for chainstring in chains:
                    cfh.write(chainstring)
                hfh.write(">" + entry + ".compressed\n" + compressedstring + "\n")
        cfh.close()

    return [compressedfasta, chainfile]

def write_assembly_bedfiles(fastaobj, args, compareparams, happrefix, bedobjects):

    # write a "genome bed file" for this fasta object, naming it <happrefix>.genome.bed
    testoutputdict = {"testgenomebed": args.prefix + "/" + happrefix + ".genome.bed"}
    testbeddict = { }
    write_test_genome_bedfile(fastaobj, args, testoutputdict, testbeddict)
    bedobjects[happrefix + "genomeregions"] = testbeddict["testgenomeregions"]

    # write a bed file of locations of Ns
    testoutputdict["testnonnbed"] = args.prefix + "/" + happrefix + ".atgcseq.bed"
    testoutputdict["testnbed"] = args.prefix + "/" + happrefix + ".nlocs.bed"
    find_all_ns(fastaobj, args, None, testoutputdict, testbeddict, "testnregions", "testnonnregions", "testnbed", "testnonnbed")
    bedobjects[happrefix + "nregions"] = testbeddict["testnregions"]
    bedobjects[happrefix + "nonnregions"] = testbeddict["testnonnregions"]





import sys
import os
import re
import shutil
import pysam
import argparse
import logging
import importlib.resources
from pathlib import Path
from GQC import output
from GQC import plots

logger = logging.getLogger(__name__)

def check_for_R():
    if shutil.which("Rscript") is None:
        logger.warning("You don\'t seem to have Rscript in your path. Plots will not be generated")
        return 1
    return 0

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Plot comparisons of the output from multiple GQC runs, reading the run info from an input file"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('-i', '--inputfile', required=True, default=None, help='input file with plot specs for GQC output directories to be compared')
    parser.add_argument('-o', '--output', type=str, required=True, help='directory name for output files')
    parser.add_argument('-B', '--benchmark', type=str, required=False, default="truth", help='name of the assembly being used as a benchmark--should be the reference sequence in the bam file')
    parser.add_argument('-c', '--config', type=str, required=False, default="benchconfig.txt", help='path to a config file specifying locations of benchmark data files')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def read_config_data(args)->dict:
    configfile = args.config
    configpath = Path(configfile)

    configvals = {}
    if not configpath.exists():
        logger.critical("A config file must exist in the default location (benchconfig.txt) or be specified with the --config option.")
        exit(1)
    else:
        logger.info("Using resource locations from " + configfile)
        with open(configfile, "r") as cr:
            configline = cr.readline()
            while configline:
                p = re.compile(r'^([^#\s]+):+\s+(\S+)$')
                match = p.match(configline)
                if match:
                    key = match.group(1)
                    value = match.group(2)
                    configvals[key] = value
                configline = cr.readline()

        # add the resource directory location for all non-absolute paths:
        if 'resourcedir' in configvals.keys():
            resourcedir = configvals["resourcedir"]
            if (os.path.exists(resourcedir)):
                for configkey in configvals.keys():
                    if configvals[configkey][0] != "/":
                        configvals[configkey] = resourcedir + "/" + configvals[configkey]
            else:
                logger.critical("The resource directory specified in the config file as \"resourcedir\" does not exist. Please change it to the location of files from the resource tarball")
                print("The resource directory specified in the config file as \"resourcedir\" does not exist. Please change it to the location of files from the resource tarball")
                exit(1)

    return configvals

def read_input_assemblies(args)->list:
    inputfile = args.inputfile
    inputpath = Path(inputfile)

    assemblies = []
    if not inputpath.exists() or not (os.access(inputpath, os.R_OK)):
        logger.critical("The assembly input file specified with the --inputfile (-i) option must exist and be readable.")
        exit(1)
    else:
        logger.info("Will create comparison plots for assemblies listed in " + inputfile)
        with open(inputfile, "r") as fh:
            assemblyline = fh.readline()
            while assemblyline:
                if assemblyline[0] == "#":
                    assemblyline = fh.readline()
                    continue
                fields = assemblyline.rstrip().split("\t")
                if len(fields) > 4:
                    assemblies.append({'directory':fields[0], 'prefix':fields[1], 'label':fields[2], 'shortlabel':fields[3], 'color':fields[4]})
                elif len(fields) == 4:
                    assemblies.append({'directory':fields[0], 'prefix':fields[1], 'label':fields[2], 'shortlabel':fields[3]})
                else:
                    logger.critical("Invalid number of tab-delimited fields in input file. Must be three, or four if colors are specified")
                    print("Invalid number of tab-delimited fields in input file. Must be three, or four if colors are specified")
                    exit(1)
                assemblyline = fh.readline()

    return assemblies

def main() -> None:

    args = parse_arguments(sys.argv[1:])
    no_rscript = check_for_R()
    
    outputdir = args.output
    outputdir.strip('/')
    logfile = outputdir + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    outputdirname = output.create_output_directory(outputdir)

    benchparams = read_config_data(args)
    inputassemblies = read_input_assemblies(args)

    # plot input file's directory results
    plots.plot_assembly_comparison_plots(inputassemblies, args.benchmark, benchparams["nonnseq"], outputdirname)


if __name__ == "__main__":
    main()



import sys
import os
import re
import shutil
import pysam
import argparse
import logging
import pybedtools
import importlib.resources
from pathlib import Path
from GQC import seqparse
from GQC import alignparse
from GQC import bedtoolslib
from GQC import heatmap

logger = logging.getLogger(__name__)

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Print all heterozygous variants in a specified set of assembly alignments"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('--bam1', required=True, help='bam file of alignments in one direction from one haplotype mapped to the other')
    parser.add_argument('--bam2', required=True, help='bam file of alignments in the other direction')
    parser.add_argument('--ref1', type=str, required=True, help='(indexed) reference fasta file for the first bam file')
    parser.add_argument('--ref2', type=str, required=True, help='(indexed) reference fasta file for the second bam file')
    parser.add_argument('-m', '--minalignlength', type=int, required=False, default=10000, help='minimum length of alignment required to be included in het site gathering')
    parser.add_argument('--windowsize', '--hetwindowsize', type=int, required=False, default=100000, help='window size for reporting heterozygosity levels')
    parser.add_argument('-n', '--non1to1', action='store_true', required=False, help='relax 1to1 (default) requirement of matching alignment endpoints')
    parser.add_argument('--heatmap', action='store_true', required=False, help='print a bed file with windowed variant counts and colors representing heterozygosity levels (red high yellow low)')
    parser.add_argument('--heatmapmincount', type=int, required=False, help='window variant count to correspond to the yellow end of the heatmap spectrum')
    parser.add_argument('--heatmapmaxcount', type=int, required=False, help='window variant count to correspond to the red end of the heatmap spectrum')
    parser.add_argument('-g', '--gfa', type=str, required=False, help='gfa file for assembly graph to guide selection of het alignments')
    parser.add_argument('-p', '--prefix', type=str, required=False, default="hetsites", help='prefix to use in output filenames')
    parser.add_argument('-s', '--splitchar', type=str, required=False, default="_", help='delimiter in scaffold names separating autosome names to be compared between haplotypes')
    parser.add_argument('-c', '--chromindex', type=int, required=False, default=0, help='index of autosome name in scaffold name when split on the split character')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging purposes')

    return parser

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    return args

def check_for_bedtools():
    if shutil.which("bedtools") is None:
        print("You don\'t seem to have bedtools in your path. Please install bedtools")
        logger.critical("You don\'t seem to have bedtools in your path. Please install bedtools")
        exit(1)
    return 0

# reads aligns from a bam file, and returns (a) a BED object with all same-autosome (or X/Y) aligned reference intervals (unmerged)
# (b) a BED object with query intervals, and (c) a dictionary of alignments query-able by alignment name (which is query:start-end/ref:start-end
def read_aligns(bamobj, args):
    aligns = {}

    querybedstring = ''
    refbedstring = ''
    #alignnumber = 1
    for align in bamobj.fetch():
        if align.is_secondary:
            continue
        if align.reference_length >= args.minalignlength:
            query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
            querycoords = query + ":" + str(querystart) + "-" + str(queryend)
            refcoords = ref + ":" + str(refstart) + "-" + str(refend)
            querychrom = query.split(args.splitchar)[args.chromindex]
            refchrom = ref.split(args.splitchar)[args.chromindex]

            if querychrom == refchrom or (querychrom == "chrX" and refchrom == "chrY") or (querychrom == "chrY" and refchrom == "chrX"):
                alignname = querycoords + "/" + refcoords
                aligns[alignname] = align
                querybedstring = querybedstring + query + "\t" + str(querystart) + "\t" + str(queryend) + "\t" + alignname + "\n"
                refbedstring = refbedstring + ref + "\t" + str(refstart) + "\t" + str(refend) + "\t" + alignname + "\n"

    refalignintervals = pybedtools.BedTool(refbedstring, from_string=True)
    queryalignintervals = pybedtools.BedTool(querybedstring, from_string=True)

    return refalignintervals, queryalignintervals, aligns

def find_corresponding_alignment_pairs(alignobj1, alignobj2, refobj1, refobj2, aligndict1, aligndict2, querybed1, querybed2, args):

    alignmapping = {}
    orthologymapstring1 = ""
    orthologymapstring2 = ""
    # look at aligns in the first direction, e.g., maternal aligned to paternal
    for align1 in alignobj1.fetch():
        query1, querystart1, queryend1, ref1, refstart1, refend1, strand1 = alignparse.retrieve_align_data(align1)
        refcoords1 = ref1 + ":" + str(refstart1) + "-" + str(refend1)
        querycoords1 = query1 + ":" + str(querystart1) + "-" + str(queryend1)
        align1name = querycoords1+"/"+refcoords1
        align2name = refcoords1+"/"+querycoords1

        # when there are reciprocal best alignments between haplotypes with matching endpoints in bam1 and bam2, use them:
        if align1name in aligndict1.keys() and align2name in aligndict2.keys():
            align2 = aligndict2[align2name]
            logger.debug("MATCH\t" + align1name + " and " + align2name)
            query2, querystart2, queryend2, ref2, refstart2, refend2, strand2 = alignparse.retrieve_align_data(align2)
            refcoords2 = ref2 + ":" + str(refstart2) + "-" + str(refend2)
            if querycoords1 == refcoords2 and refend1 - refstart1 >= args.minalignlength:
                alignmapping[align1name] = {'align':align1, 'intervals':None}
                alignmapping[align2name] = {'align':align2, 'intervals':None}
                orthologymapstring1 = orthologymapstring1 + ref1 + "\t" + str(refstart1 - 1) + "\t" + str(refend1) + "\t" + align1name + ".1to1" + "\n"
                orthologymapstring2 = orthologymapstring2 + ref2 + "\t" + str(refstart2 - 1) + "\t" + str(refend2) + "\t" + align2name + ".1to1" + "\n"
            elif refend1 - refstart1 >= args.minalignlength:
                logger.warn("Matching alignment " + align2name + " has unexpected reference coords")
        # without 1-1 requirement, will also allow alignments contained in larger opposite-direction alignments to be included
        elif args.non1to1 and align1name in aligndict1.keys():
            refinterval = pybedtools.BedTool(ref1 + "\t" + str(refstart1) + "\t" + str(refend1), from_string = True)
            intersectingaligns = bedtoolslib.intersectintervals(refinterval, querybed2, wb=True)
            for aligninterval in intersectingaligns:
                align2name = aligninterval[6]
                align2chrom = aligninterval.chrom
                align2start = aligninterval.start
                align2end = aligninterval.end
                logger.debug(align1name + " intersectalign " + align2name)
                querycoords2 = aligninterval.chrom + ":" + str(aligninterval.start) + "-" + str(aligninterval.end)
                if align2name in aligndict2.keys():
                    align2 = aligndict2[align2name]
                    query2, querystart2, queryend2, ref2, refstart2, refend2, strand2 = alignparse.retrieve_align_data(align2)
                    querycoords2 = query2 + ":" + str(querystart2) + "-" + str(queryend2)
                    refcoords2 = ref2 + ":" + str(refstart2) + "-" + str(refend2)
                    # trim to common within endpoints from bam1:
                    [commonref1start, commonref1end] = alignparse.compare_alignments(align1, align2, switched=True)
                    logger.debug("MISMATCH\t" + refcoords1 + "/" + querycoords1 + "\t" + str(commonref1start) + "\t" + str(commonref1end))
                    if commonref1start is not None and commonref1end is not None and commonref1start != -1:
                        if align1name not in alignmapping.keys():
                            alignmapping[align1name] = {'align':align1, 'intervals':[[commonref1start, commonref1end]]}
                        elif alignmapping[align1name]['intervals'] is None:
                            logger.debug("Align " + align1name + " already exists in alignmapping dictionary with no restriction intervals")
                            alignmapping[align1name]['intervals'] = [[commonref1start, commonref1end]]
                        else:
                            alignmapping[align1name]['intervals'].append([commonref1start, commonref1end])
                        orthologymapstring1 = orthologymapstring1 + ref1 + "\t" + str(refstart1 - 1) + "\t" + str(refend1) + "\t" + align1name + "." + str(commonref1start) + "_" + str(commonref1end) + "\n"
                    elif commonref1start is None and queryend1 - querystart1 > 1000000:
                        logger.debug("LONG MISMATCH length " + str(queryend1 - querystart1))
                    # now trim to common within endpoints from bam2:
                    [commonref2start, commonref2end] = alignparse.compare_alignments(align2, align1, switched=True)
                    logger.debug("MISMATCH\t" + refcoords2 + "/" + querycoords2 + "\t" + str(commonref2start) + "\t" + str(commonref2end))
                    if commonref2start is not None and commonref2end is not None and commonref2start != -1:
                        if align2name not in alignmapping.keys():
                            alignmapping[align2name] = {'align':align2, 'intervals':[[commonref2start, commonref2end]]}
                        elif alignmapping[align2name]['intervals'] is None:
                            logger.debug("Align " + align2name + " already exists in alignmapping dictionary with no restriction intervals")
                            alignmapping[align1name]['intervals'] = [[commonref1start, commonref1end]]
                        else:
                            alignmapping[align2name]['intervals'].append([commonref2start, commonref2end])
                        orthologymapstring2 = orthologymapstring2 + ref2 + "\t" + str(refstart2 - 1) + "\t" + str(refend2) + "\t" + align2name + "." + str(commonref2start) + "_" + str(commonref2end) + "\n"
                    elif commonref2start is None and queryend2 - querystart2 > 1000000:
                        logger.debug("LONG MISMATCH length " + str(queryend2 - querystart2))

    with open(args.prefix + ".orthoregions1.bed", "w") as ofh1:
        ofh1.write(orthologymapstring1)

    with open(args.prefix + ".orthoregions2.bed", "w") as ofh2:
        ofh2.write(orthologymapstring2)

    return alignmapping

def find_hets_and_coveredregions(bamobj, refobj, queryobj, alignmap:dict, args):

    # strings which will be used to create bed files with pybedtools:
    aligncoveredregions = ""
    aligncoveredwindows = ""
    hetvariants = ""
    # to avoid duplicates due to overlapping alignments:
    includedvariantdict = {}
    allwindowintervalswithcounts = pybedtools.BedTool("", from_string=True)
    for align in bamobj.fetch():
        query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
        querycoords = query + ":" + str(querystart) + "-" + str(queryend)
        refcoords = ref + ":" + str(refstart) + "-" + str(refend)
        alignname = querycoords + "/" + refcoords

        windowintervalswithcounts = pybedtools.BedTool("", from_string=True)
        boundarylist = []
        if alignname in alignmap.keys():
            if alignmap[alignname]['intervals'] is None:
                boundarylist.append([refstart, refend])
            else:
                for intervalpair in alignmap[alignname]['intervals']:
                    boundarylist.append(intervalpair)

            for intervalpair in boundarylist:
                [restrictedstart, restrictedend] = intervalpair
                # add query\tstart\tstop\tname\tscore\tstrand to covered string:
                if restrictedstart < 1 or restrictedend < 1:
                    logger.debug("Ignoring alignment from " + querycoords + " to " + refcoords + " due to invalid restrictedstart " + str(restrictedstart))
                else:
                    coveredstring = ref + "\t" + str(restrictedstart - 1) + "\t" + str(restrictedend) + "\t" + querycoords + "/" + strand + "\t0.0\t" + strand + "\n"
                    aligncoveredregions = aligncoveredregions + coveredstring
                    coveredwindows = make_covered_window_string(ref, restrictedstart, restrictedend, query, querystart, queryend, strand, args)
        
                    aligncoveredwindows = aligncoveredwindows + coveredwindows
    
                    alignhetvariants = ""
                    novelhetvariantstring = ""
                    variantlist = alignparse.align_variants(align, queryobj, query, querystart, queryend, refobj, ref, refstart, refend, strand)
                    for variant in variantlist:
                        chrom = variant.chrom
                        start = variant.start
                        end = variant.end
                        if start >= restrictedstart and end < restrictedend:
                            name = variant.name
                            alignment = alignname + "." + str(restrictedstart) + "_" + str(restrictedend)
                            variantstring = chrom + "\t" + str(start) + "\t" + str(end) + "\t" + name + "\t" + alignment + "\n"
                            alignhetvariants = alignhetvariants + variantstring
                            if name not in includedvariantdict.keys():
                                novelhetvariantstring = novelhetvariantstring + variantstring
                                includedvariantdict[name] = 1
    
                    hetvariants = hetvariants + novelhetvariantstring
    
                    variantintervals = pybedtools.BedTool(alignhetvariants, from_string=True)
                    coveredwindowintervals = pybedtools.BedTool(coveredwindows, from_string=True)
    
                    alignintervalswithcounts = bedtoolslib.intersectintervals(coveredwindowintervals, variantintervals, wa=True, counts=True)

                    numvars = len(variantintervals)
                    numwindows = len(coveredwindowintervals)
                    numcounts = len(alignintervalswithcounts)
                    logger.debug("Intersected " + str(numwindows) + " windows with " + str(numvars) + " variants and got " + str(numcounts) + " intervals with counts for align " + alignname)
                    windowintervalswithcounts = windowintervalswithcounts.cat(alignintervalswithcounts, postmerge=False)
                    logger.debug("windowintervalwithcounts has " + str(len(windowintervalswithcounts)) + " windows with counts for align " + alignname)

        logger.debug("Gathered " + str(len(windowintervalswithcounts)) + " windows with counts for align " + alignname)
        allwindowintervalswithcounts = allwindowintervalswithcounts.cat(windowintervalswithcounts, postmerge=False)

    return aligncoveredregions, allwindowintervalswithcounts, hetvariants

def make_covered_window_string(ref:str, refstart:int, refend:int, query:str, querystart:int, queryend:int, strand:str, args):
    windowsize = args.windowsize

    # if intervals is smaller than the windowsize, just return the interval as the window:
    if windowsize > refend - refstart:
        windowname = ref + "_" + str(refstart) + "_" + str(refend) + "_" + query + "_" + str(querystart) + "_" + str(queryend) + "_" + strand + "_0"
        bedstring = ref + "\t" + str(refstart) + "\t" + str(refend) + "\t" + windowname + "\n"
        return bedstring

    # otherwise return windows starting at "nice" positions with flanking partial windows:
    windowstart = windowsize*int(refstart/windowsize + 1.0)
    if windowstart - refstart > 0:
        bedstring = ref + "\t" + str(refstart) + "\t" + str(windowstart) + "\t" + ref + "_" + str(refstart) + "_" + str(refend) + "_" + query + "_" + str(querystart) + "_" + str(queryend) + "_" + strand + "_0" + "\n"
    else:
        bedstring = ""
    lastwindowend = windowsize*int(refend/windowsize)
    if refend < lastwindowend:
        lastwindowend = refend

    windownumber = 1
    while windowstart + windowsize <= lastwindowend:
        windowname = ref + "_" + str(refstart) + "_" + str(refend) + "_" + query + "_" + str(querystart) + "_" + str(queryend) + "_" + strand + "_" + str(windownumber)
        bedstring = bedstring + ref + "\t" + str(windowstart) + "\t" + str(windowstart + windowsize) + "\t" + windowname + "\n"
        windownumber = windownumber + 1
        windowstart = windowstart + windowsize

    if windowstart < refend:
        windowname = ref + "_" + str(refstart) + "_" + str(refend) + "_" + query + "_" + str(querystart) + "_" + str(queryend) + "_" + strand + "_" + str(windownumber)
        bedstring = bedstring + ref + "\t" + str(windowstart) + "\t" + str(refend) + "\t" + windowname + "\n"

    return bedstring

def main() -> None:

    args = parse_arguments(sys.argv[1:])

    logfile = args.prefix + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    # make sure user has bedtools:
    check_for_bedtools()

    # read aligns from BAM files
    # coords dictionaries have keys that are the query coordinates and values that are the ref coordinates
    if args.bam1 is not None:
        alignobj1 = pysam.AlignmentFile(args.bam1, "rb")
        refbed1, querybed1, aligndict1 = read_aligns(alignobj1, args)
        querybed1.sort().saveas(args.prefix + ".hetquery1.bed")
    else:
        alignobj1 = None
    if args.bam2 is not None:
        alignobj2 = pysam.AlignmentFile(args.bam2, "rb")
        refbed2, querybed2, aligndict2 = read_aligns(alignobj2, args)
        querybed2.sort().saveas(args.prefix + ".hetquery2.bed")
    else:
        alignobj2 = None
    # refs
    if args.ref1 is not None:
        refobj1 = pysam.FastaFile(args.ref1)
    if args.ref2 is not None:
        refobj2 = pysam.FastaFile(args.ref2)

    # should we parse a gfa file? This is just an experimental option--not a true part of the program (random, I know)
    if args.gfa is not None:
        genomegraph = assemblygraph.read_graph_from_gfa(args.gfa, args)
        assemblygraph.find_hom_het_nodes(genomegraph, args)
        [compressed_ref1, chainfile1] = seqparse.compress_sequence(args.ref1, args)
        [compressed_ref2, chainfile2] = seqparse.compress_sequence(args.ref2, args)

        compare_compressed_fasta_to_node_seqs(compressed_ref1, args.gfa, args)
        compare_compressed_fasta_to_node_seqs(compressed_ref2, args.gfa, args)

    # look for corresponding alignments in both directions:
    if alignobj1 is not None:
        alignmapping = find_corresponding_alignment_pairs(alignobj1, alignobj2, refobj1, refobj2, aligndict1, aligndict2, querybed1, querybed2, args)
       
        coveredregions1, windowswithcounts1, hetvariants1 = find_hets_and_coveredregions(alignobj1, refobj1, refobj2, alignmapping, args)
        coveredregions2, windowswithcounts2, hetvariants2 = find_hets_and_coveredregions(alignobj2, refobj2, refobj1, alignmapping, args)

        logger.debug("find_hets_and_coveredregion returned " + str(len(windowswithcounts1)) + " and " + str(len(windowswithcounts2)) + " windows with counts")
        allvars = hetvariants1 + hetvariants2
        variantintervals = pybedtools.BedTool(allvars, from_string=True)
        variantintervalssorted = variantintervals.sort()
        #mergedvariants = bedtoolslib.mergeintervals(variantintervalssorted, collapsecolumns='4,5', collapseoutput='distinct', collapsedelim=',', distance=-1)
        variantintervalssorted.saveas(args.prefix + ".hetsites.bed")
        allcovered = coveredregions1 + coveredregions2
        coveredintervals = pybedtools.BedTool(allcovered, from_string=True)
        coveredintervals.sort().saveas(args.prefix + ".hapaligned.bed")
        windowwithcountintervals = windowswithcounts1.cat(windowswithcounts2, postmerge=False)
        windowwithcountintervals.sort().saveas(args.prefix + ".hapaligned.windows.withcounts.bed")
    
        if args.heatmap:
            heatmapwindows = heatmap.add_colors_to_intervals(windowwithcountintervals, args)
            if heatmapwindows is not None:
                heatmapwindows.sort().saveas(args.prefix + ".hapaligned.windows.withheatmap.bed")

if __name__ == "__main__":
    main()



import sys
import os
import re
import shutil
import pysam
import argparse
import pybedtools
import logging
import importlib.resources
from pathlib import Path
from GQC import errors
from GQC import output
from GQC import seqparse
from GQC import alignparse
from GQC import phasing
from GQC import coverage
from GQC import stats
from GQC import plots

logger = logging.getLogger(__name__)

def check_for_R():
    if shutil.which("Rscript") is None:
        logger.warning("You don\'t seem to have Rscript in your path. Plots will not be generated")
        return 1
    return 0

def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] [FILE]...",
        description="Print read quality statistics from a bam file of the reads aligned to a benchmark assembly"
    )
    parser.add_argument(
        "-v", "--version", action="version",
        version = f"{parser.prog} version 0.1.0"
    )
    parser.add_argument('-b', '--bam', required=True, default=None, help='bam file of alignments of the reads to the diploid benchmark')
    parser.add_argument('-r', '--reffasta', type=str, required=True, help='(indexed) fasta file for benchmark reference')
    parser.add_argument('--regions', '--regionbed', type=str, required=False, default='', help='bed file of benchmark regions to assess (default is to assess the whole diploid benchmark genome)')
    parser.add_argument('-p', '--prefix', type=str, required=True, help='prefix for output directory name')
    parser.add_argument('--mincontiglength', type=int, required=False, default=500, help='minimum length for contig to be included in contig statistics')
    parser.add_argument('--maxstrreads', type=int, required=False, default=500, help='maximum number of reads to be evaluated for any one STR run in the STR run accuracy analysis. Use 0 to analyze all reads')
    parser.add_argument('--excludefile', type=str, required=False, default=None, help='bed file of benchmark locations to exclude from consideration (in addition to stretches of 10 or more Ns and regions in the exclude file specified in the config file)')
    parser.add_argument('--strs', action='store_true', required=False, help='perform analysis of short tandem repeat length accuracy')
    parser.add_argument('--baseerrors', action='store_true', required=False, help='perform analysis of base errors within reads')
    parser.add_argument('--bincoverage', action='store_true', required=False, help='perform analysis of binned read coverage')
    parser.add_argument('--kmercoverage', action='store_true', required=False, help='perform analyses of read kmer coverage')
    parser.add_argument('--arrivalratecoverage', action='store_true', required=False, help='perform analyses of read arrival rates (for test against Poisson)')
    parser.add_argument('--downsample', type=restricted_float, required=False, default=None, help='fraction of read alignments to include in error reporting statistics calculations (must be a floating point number between 0 and 1)')
    parser.add_argument('--covbinsize', type=int, required=False, default=0, help='size of bins used to tally read counts in coverage analysis. If 0, will calculate bin size to result in roughly 1000 read starts per bin.')
    parser.add_argument('--bincovoverlap', action='store_true', required=False, help='count reads that overlap bins, rather than just reads that start in bins')
    parser.add_argument('--covkmersize', type=int, required=False, default=3, help='size of kmers used in coverage analysis. Values greater than 5 will cause only "extreme" kmers composed of two bases to be analyzed.')
    parser.add_argument('--minreadalignedpercentage', type=int, required=False, default=90, help='In the coverage analysis, minimum percentage of read required to be aligned without clipping.')
    parser.add_argument('-e', '--errorfile', type=str, required=False, default='', help='preexisting file of read errors to report and plot stats for')
    parser.add_argument('-R', '--readsetname', type=str, required=False, default="test", help='name of the assembly being tested--should be query sequence in bam file')
    parser.add_argument('-B', '--benchmark', type=str, required=False, default="truth", help='name of the assembly being used as a benchmark--should be the reference sequence in the bam file')
    parser.add_argument('-c', '--config', type=str, required=False, default="benchconfig.txt", help='path to a config file specifying locations of benchmark data files')
    parser.add_argument('--rerun', action='store_true', required=False, help='use existing file of read errors rather than recreating it')
    parser.add_argument('--debug', action='store_true', required=False, help='print verbose output to log file for debugging')

    return parser

# function to check that argument is within range from 0 to 1
def restricted_float(x):
    try:
        x = float(x)
    except ValueError:
        raise argparse.ArgumentTypeError("%r not a floating-point literal" % (x,))

    if x <= 0.0 or x > 1.0:
        raise argparse.ArgumentTypeError("%r not in range (0.0, 1.0]"%(x,))
    return x

def parse_arguments(args):
    parser = init_argparse()
    args = parser.parse_args(args)

    if not args.bam:
        logger.critical("Must specify a bam file with --bam")
        exit(1)

    return args

def read_config_data(args)->dict:
    configfile = args.config
    configpath = Path(configfile)

    configvals = {}
    if not configpath.exists():
        logger.critical("A config file must exist in the default location (benchconfig.txt) or be specified with the --config option.")
        exit(1)
    else:
        logger.info("Using resource locations from " + configfile)
        with open(configfile, "r") as cr:
            configline = cr.readline()
            while configline:
                p = re.compile(r'^([^#\s]+):+\s+(\S+)$')
                match = p.match(configline)
                if match:
                    key = match.group(1)
                    value = match.group(2)
                    configvals[key] = value
                configline = cr.readline()

        # add the resource directory location for all non-absolute paths:
        if 'resourcedir' in configvals.keys():
            resourcedir = configvals["resourcedir"]
            if (os.path.exists(resourcedir)):
                for configkey in configvals.keys():
                    if configvals[configkey][0] != "/":
                        configvals[configkey] = resourcedir + "/" + configvals[configkey]
            else:
                logger.critical("The resource directory specified in the config file as \"resourcedir\" does not exist. Please change it to the location of files from the resource tarball")
                print("The resource directory specified in the config file as \"resourcedir\" does not exist. Please change it to the location of files from the resource tarball")
                exit(1)

    return configvals

def main() -> None:

    args = parse_arguments(sys.argv[1:])
    #check_for_bedtools()
    no_rscript = check_for_R()
    
    logfile = args.prefix + ".log"
    logformat = '%(asctime)s %(message)s'
    if args.debug:
        logging.basicConfig(filename=logfile, level=logging.DEBUG, format=logformat)
        logger.info('Logging verbose output for debugging.')
    else:
        logging.basicConfig(filename=logfile, level=logging.INFO, format=logformat)

    outputdir = output.create_output_directory(args.prefix)
    benchparams = read_config_data(args)

    hetsites = phasing.read_hetsites(benchparams["hetsitevariants"])

    alignobj = pysam.AlignmentFile(args.bam, "rb")
    refobj = pysam.FastaFile(args.reffasta)

    outputfiles = output.name_read_stats_files(args, outputdir)

    benchintervals = seqparse.write_included_bedfile(refobj, args, benchparams, outputfiles)

    if args.downsample is not None:
        logger.info("Downsampling to fraction " + str(args.downsample) + " of reads")

    if args.bincoverage:
        logger.info("Calculating binned coverage")
        logger.debug(outputfiles["coveragebedfile"])
        coverage.tally_bin_coverages(alignobj, refobj, benchintervals, outputfiles["includedbedfile"], outputfiles["coveragebedfile"], outputfiles["includedcoveragebedfile"], args)
        #plots.plot_read_coverage_vs_gccontent(outputfiles["coveragebedfile"], outputfiles["extremekmersbedfile"])

    if args.arrivalratecoverage:
        logger.info("Calculating read arrival rate in bins for comparison to Poisson distribution")
        logger.debug(outputfiles["arrivalratebedfile"])
        coverage.tally_included_bin_arrival_rates(alignobj, refobj, benchintervals, outputfiles["arrivalratebedfile"], args)

    if args.kmercoverage:
        coverage.compare_read_kmers_to_benchmark_kmers(alignobj, refobj, benchintervals, outputfiles["benchmarkkmercountfile"], outputfiles["readalignedkmercountfile"], outputfiles["strandedalignedkmercountfile"], benchparams, args)

    if args.strs:
        logger.info("Assessing accuracy of short tandem repeats")
        print("Assessing accuracy of short tandem repeats")
    
        if "tetranucruns" in benchparams:
            logger.debug(benchparams["tetranucruns"])
            tetranucstats = errors.assess_str_read_coverage_require_flank('tetranuc', alignobj, refobj, benchparams["tetranucruns"], outputfiles, benchintervals, hetsites, args)
            stats.write_read_str_stats('tetranuc', tetranucstats, outputfiles, args)
        else:
            logger.info("No tetranucleotide bed file specified in configuration file(tetranucruns)")
        
        if "trinucruns" in benchparams:
            logger.debug(benchparams["trinucruns"])
            trinucstats = errors.assess_str_read_coverage_require_flank('trinuc', alignobj, refobj, benchparams["trinucruns"], outputfiles, benchintervals, hetsites, args)
            stats.write_read_str_stats('trinuc', trinucstats, outputfiles, args)
        else:
            logger.info("No trinucleotide bed file specified in configuration file(trinucruns)")
    
        if "dinucruns" in benchparams:
            logger.debug(benchparams["dinucruns"])
            dinucstats = errors.assess_str_read_coverage_require_flank('dinuc', alignobj, refobj, benchparams["dinucruns"], outputfiles, benchintervals, hetsites, args)
            stats.write_read_str_stats('dinuc', dinucstats, outputfiles, args)
        else:
            logger.info("No dinucleotide bed file specified in configuration file(dinucruns)")
    
        if "mononucruns" in benchparams:
            logger.debug(benchparams["mononucruns"])
            mononucstats = errors.assess_str_read_coverage_require_flank('mononuc', alignobj, refobj, benchparams["mononucruns"], outputfiles, benchintervals, hetsites, args)
            stats.write_read_str_stats('mononuc', mononucstats, outputfiles, args)
        else:
            logger.info("No mononucleotide bed file specified in configuration file(mononucruns)")

    # evaluate errors within read alignments:
    if args.baseerrors:
        logger.info("Assessing errors within read alignments")
        logger.debug(outputfiles["readerrorfile"])
        errorstats = errors.assess_read_align_errors(alignobj, refobj, outputfiles["readerrorfile"], benchintervals, hetsites, args)
        stats.write_read_error_summary(errorstats, outputfiles)
        if len(errorstats["alignedqualscorecounts"]) > 0:
            plots.plot_read_error_stats(args.readsetname, args.benchmark, outputdir)


if __name__ == "__main__":
    main()
